import numpy as np, sys, os, random, pdb, json, uuid, time, argparse
from pprint import pprint
import logging, logging.config
from collections import defaultdict as ddict
# from ordered_set import OrderedSet

# PyTorch related imports
import torch
from torch.nn import functional as F
from torch.nn.init import xavier_normal_
from torch.utils.data import DataLoader
from torch.nn import Parameter
from torch_scatter import scatter_add

np.set_printoptions(precision=4)

def set_gpu(gpus):
	"""
	Sets the GPU to be used for the run

	Parameters
	----------
	gpus:           List of GPUs to be used for the run
	
	Returns
	-------
		
	"""
	os.environ["CUDA_DEVICE_ORDER"]    = "PCI_BUS_ID"
	os.environ["CUDA_VISIBLE_DEVICES"] = gpus

def get_logger(name, log_dir, config_dir):
	"""
	Creates a logger object

	Parameters
	----------
	name:           Name of the logger file
	log_dir:        Directory where logger file needs to be stored
	config_dir:     Directory from where log_config.json needs to be read
	
	Returns
	-------
	A logger object which writes to both file and stdout
		
	"""
	config_dict = json.load(open( config_dir + 'log_config.json'))
	config_dict['handlers']['file_handler']['filename'] = log_dir + name.replace('/', '-')
	logging.config.dictConfig(config_dict)
	logger = logging.getLogger(name)

	std_out_format = '%(asctime)s - [%(levelname)s] - %(message)s'
	consoleHandler = logging.StreamHandler(sys.stdout)
	consoleHandler.setFormatter(logging.Formatter(std_out_format))
	logger.addHandler(consoleHandler)

	return logger

def get_combined_results(left_results, right_results):
	results = {}
	count   = float(left_results['count'])

	results['left_mr']	= round(left_results ['mr'] /count, 5)
	results['left_mrr']	= round(left_results ['mrr']/count, 5)
	results['right_mr']	= round(right_results['mr'] /count, 5)
	results['right_mrr']	= round(right_results['mrr']/count, 5)
	results['mr']		= round((left_results['mr']  + right_results['mr']) /(2*count), 5)
	results['mrr']		= round((left_results['mrr'] + right_results['mrr'])/(2*count), 5)

	for k in range(10):
		results['left_hits@{}'.format(k+1)]	= round(left_results ['hits@{}'.format(k+1)]/count, 5)
		results['right_hits@{}'.format(k+1)]	= round(right_results['hits@{}'.format(k+1)]/count, 5)
		results['hits@{}'.format(k+1)]		= round((left_results['hits@{}'.format(k+1)] + right_results['hits@{}'.format(k+1)])/(2*count), 5)
	return results

def get_param(shape):
	param = Parameter(torch.Tensor(*shape)); 	
	xavier_normal_(param.data)
	return param

def com_mult(a, b):
	r1, i1 = a[..., 0], a[..., 1]
	r2, i2 = b[..., 0], b[..., 1]
	return torch.stack([r1 * r2 - i1 * i2, r1 * i2 + i1 * r2], dim = -1)

def conj(a):
	a[..., 1] = -a[..., 1]
	return a

def cconv(a, b):
	return torch.irfft(com_mult(torch.rfft(a, 1), torch.rfft(b, 1)), 1, signal_sizes=(a.shape[-1],))

def ccorr(a, b):
	return torch.irfft(com_mult(conj(torch.rfft(a, 1)), torch.rfft(b, 1)), 1, signal_sizes=(a.shape[-1],))



import json
import os

import torch
from torch.autograd import Variable

from .mask_generation import GenerateGraphMask


def get_normalized_masked_ids(masked_node_ids, ent2id):
    """
    Input : [batch * list[masked_node_id]]
    returns : [batch * list(normalized_node_id) ] #len(attr-graph.get_number_of_nodes)]
    """

    normalized_node_ids = list()
    for batch_node_ids in masked_node_ids:
        normalized_node_ids_batch = [
            ent2id[str(x.cpu().numpy().tolist())] for x in batch_node_ids
        ]
        normalized_node_ids.append(normalized_node_ids_batch)

    return torch.LongTensor(normalized_node_ids)


class Processing_GCN_Walks:
    def __init__(self, nodeid2rowid, relations, n_pred, max_length, max_pred):

        self.nodeid2rowid = nodeid2rowid
        self.relations = relations
        self.n_pred = n_pred
        self.max_length = max_length
        self.max_pred = max_pred
        self.special_tokens = {}

    def getSpecialTokens(self):
        no_nodes = 0
        for node_type in self.nodeid2rowid:
            no_nodes += len(self.nodeid2rowid[node_type])
        special_tokens_set = ["[PAD]", "[MASK]"]
        for ii, _ in enumerate(special_tokens_set):
            self.special_tokens[special_tokens_set[ii]] = no_nodes + ii

        return self.special_tokens

    def masking_and_padding(self, subgraphs_list):
        for ii, _ in enumerate(subgraphs_list):
            # each subgraph has max_length edges
            subgraphs_list[ii] = subgraphs_list[ii][: self.max_length + 1]

        # generate mask
        graph_mask = GenerateGraphMask(subgraphs_list, self.n_pred)
        input_mask, masked_nodes, masked_position = graph_mask.GCN_MaskGeneration()
        # print('\n input mask:')
        # for ii in range(len(input_mask)):
        #     print(input_mask[ii])

        for ii, _ in enumerate(subgraphs_list):
            for jj in range(len(subgraphs_list[ii])):
                if input_mask[ii][jj] == 0:
                    subgraphs_list[ii][jj] = self.special_tokens["[MASK]"]
            # padding subgraph to max_length
            if len(subgraphs_list[ii]) < self.max_length + 1:
                n_pad = self.max_length + 1 - len(subgraphs_list[ii])
                for _ in range(n_pad):
                    subgraphs_list[ii].append(self.special_tokens["[PAD]"])

        return subgraphs_list, masked_nodes, masked_position

    def process_minibatch(self, data_path, data_name, task, batch_id):
        folder = os.path.join(data_path, task)
        file_name = data_name + "_" + task + "_batch_" + str(batch_id) + ".txt"
        fp = open(os.path.join(folder, file_name), "r")
        subgraphs = []
        cnt = 0
        for line in fp:
            line = json.loads(line)
            subgraphs.append(line)
            cnt += 1
            # if cnt == 3:
            #     break
        fp.close()

        all_nodes = []
        for subgraph in subgraphs:
            nodes = []
            for edge in subgraph:
                source_node = edge[0]
                target_node = edge[1]
                if int(source_node) not in nodes:
                    nodes.append(int(source_node))
                if int(target_node) not in nodes:
                    nodes.append(int(target_node))
            all_nodes.append(nodes)

        special_tokens = self.getSpecialTokens()
        all_nodes, masked_nodes, masked_postion = self.masking_and_padding(all_nodes)
        # print('\n subgraphs \n', subgraphs)
        # print('\n all_nodes \n', all_nodes)
        # print('\n masked_nodes\n', masked_nodes)
        # print('\n masked_postion\n', masked_postion)

        masked_nodes = Variable(torch.LongTensor(masked_nodes))
        masked_postion = Variable(torch.LongTensor(masked_postion))

        return subgraphs, all_nodes, masked_nodes, masked_postion

    def pad_max_seq_len(self, subgraphs_list, max_seq_len):
        padded_subgraphs = []
        for subgraph in subgraphs_list:
            subgraph_actual_len = len(subgraph)
            for _ in range(max_seq_len - subgraph_actual_len):
                subgraph.append(self.special_tokens["[PAD]"])
            padded_subgraphs.append(subgraph)

        return padded_subgraphs

    def process_finetune_minibatch(self, finetune_path, data_name, task, batch_id):
        """
        loads all subgraph for a given batch = batch_id , from
        finetune_path + "/" task + "/" + data_name+'_'+task+'_batch_'+str(batch_id)+'.txt'
        Expects each line of format
        [[source, target, relation, 0/1]]
        The last value indicate true or false edge
        """
        folder = os.path.join(finetune_path, task)
        file_name = data_name + "_" + task + "_batch_" + str(batch_id) + ".txt"
        fp = open(os.path.join(folder, file_name), "r")
        subgraphs = []
        labels = []
        cnt = 0
        for line in fp:
            line = json.loads(line)
            labels.append(int(line[0][-1]))
            line[0] = line[0][:-1]
            subgraphs.append(line)
            cnt += 1
            # if cnt == 3:
            #     break
        fp.close()

        # print("In fine tuning process batch, loaded walks", all_walks)
        all_nodes = []
        for subgraph in subgraphs:
            nodes = []
            for edge in subgraph:
                source_node = edge[0]
                target_node = edge[1]
                if source_node not in nodes:
                    nodes.append(source_node)
                if target_node not in nodes:
                    nodes.append(target_node)
            all_nodes.append(nodes)

        special_tokens = self.getSpecialTokens()
        all_nodes = self.pad_max_seq_len(all_nodes, self.max_length + 1)
        # print("created rowid subgraphs for batch", subgraphs_list, subgraph_len)
        labels = Variable(torch.FloatTensor(labels)).unsqueeze(1)
        # print(subgraphs, labels)
        #print(labels)

        return subgraphs, all_nodes, labels



from .generic_attributed_graph import *
from .context_generator import *
from .attributed_graph import *
from .process_unb_cyber_network import *
from .negative_sampling import *
from .coherent_context_generator import *


"""
    Mapping helpers
"""
import numpy as np
import pandas as pd
import scanpy as sc
import torch
import logging
import copy
from scipy.sparse.csc import csc_matrix
from scipy.sparse.csr import csr_matrix

from pathlib import Path
import argparse
import random
import matplotlib.cm as cm
import torch.nn as nn
from torch.autograd import Variable
import os
import torch.multiprocessing
from tqdm import tqdm
import time
from .utils import *
from .models import decon as mo
from sklearn import preprocessing
from sklearn.decomposition import NMF
from sklearn.decomposition import FastICA
from torch.nn.functional import softmax, cosine_similarity
logging.getLogger().setLevel(logging.INFO)

def seed_everything(seed_value):
    random.seed(seed_value)
    np.random.seed(seed_value)
    torch.manual_seed(seed_value)
    torch.cuda.manual_seed(seed_value)
    torch.cuda.manual_seed_all(seed_value)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = True

def pp_adatas(adata_sc, adata_sp, genes=None, gene_to_lowercase = True):
    """
    Pre-process AnnDatas so that they can be mapped. Specifically:
    - Remove genes that all entries are zero
    - Find the intersection between adata_sc, adata_sp and given marker gene list, save the intersected markers in two adatas
    - Calculate density priors and save it with adata_sp
    Args:
        adata_sc (AnnData): single cell data
        adata_sp (AnnData): spatial expression data
        genes (List): Optional. List of genes to use. If `None`, all genes are used.
    
    Returns:
        update adata_sc by creating `uns` `training_genes` `overlap_genes` fields 
        update adata_sp by creating `uns` `training_genes` `overlap_genes` fields and creating `obs` `rna_count_based_density` & `uniform_density` field
    """

    # remove all-zero-valued genes
    sc.pp.filter_genes(adata_sc, min_cells=1)
    sc.pp.filter_genes(adata_sp, min_cells=1)

    if genes is None:
        # Use all genes
        genes = adata_sc.var.index
               
    # put all var index to lower case to align
    if gene_to_lowercase:
        adata_sc.var.index = [g.lower() for g in adata_sc.var.index]
        adata_sp.var.index = [g.lower() for g in adata_sp.var.index]
        genes = list(g.lower() for g in genes)

    adata_sc.var_names_make_unique()
    adata_sp.var_names_make_unique()
    

    # Refine `marker_genes` so that they are shared by both adatas
    genes = list(set(genes) & set(adata_sc.var.index) & set(adata_sp.var.index))
    # logging.info(f"{len(genes)} shared marker genes.")

    adata_sc.uns["training_genes"] = genes
    adata_sp.uns["training_genes"] = genes
    logging.info(
        "{} training genes are saved in `uns``training_genes` of both single cell and spatial Anndatas.".format(
            len(genes)
        )
    )

    # Find overlap genes between two AnnDatas
    overlap_genes = list(set(adata_sc.var.index) & set(adata_sp.var.index))
    # logging.info(f"{len(overlap_genes)} shared genes.")

    adata_sc.uns["overlap_genes"] = overlap_genes
    adata_sp.uns["overlap_genes"] = overlap_genes
    logging.info(
        "{} overlapped genes are saved in `uns``overlap_genes` of both single cell and spatial Anndatas.".format(
            len(overlap_genes)
        )
    )

    # Calculate uniform density prior as 1/number_of_spots
    adata_sp.obs["uniform_density"] = np.ones(adata_sp.X.shape[0]) / adata_sp.X.shape[0]
    logging.info(
        f"uniform based density prior is calculated and saved in `obs``uniform_density` of the spatial Anndata."
    )

    # Calculate rna_count_based density prior as % of rna molecule count
    rna_count_per_spot = np.array(adata_sp.X.sum(axis=1)).squeeze()
    adata_sp.obs["rna_count_based_density"] = rna_count_per_spot / np.sum(rna_count_per_spot)
    logging.info(
        f"rna count based density prior is calculated and saved in `obs``rna_count_based_density` of the spatial Anndata."
    )




def map_cells_to_space(
    adata_sc,
    adata_sp,
    device="cpu",
    learning_rate=0.0001,
    num_epochs=2000,
    ):
    seed_value = 1234
    seed_everything(seed_value)
    if not set(["training_genes", "overlap_genes"]).issubset(set(adata_sc.uns.keys())):
        raise ValueError("Missing tangram parameters. Run `pp_adatas()`.")

    if not set(["training_genes", "overlap_genes"]).issubset(set(adata_sp.uns.keys())):
        raise ValueError("Missing tangram parameters. Run `pp_adatas()`.")

    assert list(adata_sp.uns["training_genes"]) == list(adata_sc.uns["training_genes"])

    training_genes = adata_sc.uns["training_genes"]
    if isinstance(adata_sc.X, csc_matrix) or isinstance(adata_sc.X, csr_matrix):
        S = np.array(adata_sc[:, training_genes].X.toarray(), dtype="float32",)
    elif isinstance(adata_sc.X, np.ndarray):
        S = np.array(adata_sc[:, training_genes].X.toarray(), dtype="float32",)
    else:
        X_type = type(adata_sc.X)
        logging.error("AnnData X has unrecognized type: {}".format(X_type))
        raise NotImplementedError

    if isinstance(adata_sp.X, csc_matrix) or isinstance(adata_sp.X, csr_matrix):
        G = np.array(adata_sp[:, training_genes].X.toarray(), dtype="float32")
    elif isinstance(adata_sp.X, np.ndarray):
        G = np.array(adata_sp[:, training_genes].X, dtype="float32")
    else:
        X_type = type(adata_sp.X)
        logging.error("AnnData X has unrecognized type: {}".format(X_type))
        raise NotImplementedError

    if not S.any(axis=0).all() or not G.any(axis=0).all():
        raise ValueError("Genes with all zero values detected. Run `pp_adatas()`.")

    sc_LC = adata_sc.obsm["X_umap"]
    sc_LC = np.array(sc_LC).astype(np.float32)
    location0_sc=((sc_LC[:,0]).astype(np.float32))
    location1_sc=((sc_LC[:,1]).astype(np.float32))

    xmax0 = max(location0_sc)
    xmin0 = min(location0_sc)
    xmax1 = max(location1_sc)
    xmin1 = min(location1_sc)

    len0=xmax0-xmin0
    len1=xmax1-xmin1
    sc_location=np.array([len0,len1])

    sp_LC = adata_sp.obsm["spatial"]
    sp_LC = np.array(sp_LC).astype(np.float32)
    location0_st = (sp_LC[:,0]).astype(np.float32)
    location1_st = (sp_LC[:,1]).astype(np.float32)

    xmax0 = max(location0_st)
    xmin0 = min(location0_st)
    xmax1 = max(location1_st)
    xmin1 = min(location1_st)

    len0=xmax0-xmin0
    len1=xmax1-xmin1
    st_location=np.array([len0,len1])

    sc_LC = sc_LC.reshape((1, 1, sc_LC.shape[0],2))
    sp_LC = sp_LC.reshape((1, 1, sp_LC.shape[0],2))

    S = S.astype(np.float32)
    #S = np.transpose(S)
    G = G.astype(np.float32)
    #G = np.transpose(G)
    #Snew = preprocessing.scale(S)
    #Gnew = preprocessing.scale(G)

    Snew = S.reshape((S.shape[0],1,S.shape[1]))
    Gnew = G.reshape((G.shape[0],1,G.shape[1]))

    Snew = np.transpose(Snew)
    Gnew = np.transpose(Gnew)

    Snew, Gnew = torch.from_numpy(Snew), torch.from_numpy(Gnew)
    
    #logT = preprocessing.FunctionTransformer(np.log1p)
    
    scaler1 = preprocessing.StandardScaler().fit(S)
    Snew_scale = scaler1.transform(S)
    scaler2 = preprocessing.StandardScaler().fit(G)
    Gnew_scale = scaler2.transform(G)
    
    #pca = FastICA(n_components=100,random_state=0,whiten='unit-variance')

    #Snew_scale = pca.fit_transform(Snew_scale)
    #Gnew_scale = pca.fit_transform(Gnew_scale)

    Snew_scale = Snew_scale.astype(np.float32)
    Gnew_scale = Gnew_scale.astype(np.float32)

    Snew_scale = Snew_scale.reshape((Snew_scale.shape[0],1,Snew_scale.shape[1]))
    Gnew_scale = Gnew_scale.reshape((Gnew_scale.shape[0],1,Gnew_scale.shape[1]))
    
    #Snew_scale=copy.deepcopy(S)
    #Gnew_scale=copy.deepcopy(G)

    #Snew_scale = Snew_scale.reshape((S.shape[0],1,S.shape[1]))
    #Gnew_scale = Gnew_scale.reshape((G.shape[0],1,G.shape[1]))
    
    Snew_scale = np.transpose(Snew_scale)
    Gnew_scale = np.transpose(Gnew_scale)

    Snew_scale, Gnew_scale = torch.from_numpy(Snew_scale), torch.from_numpy(Gnew_scale)
    #print(Snew_scale.shape)

    kpsc_np, kpst_np = torch.from_numpy(sc_LC), torch.from_numpy(sp_LC)
    sc_location, st_location = torch.from_numpy(sc_location), torch.from_numpy(st_location)
    #seed = 1
    #seed_everything(seed)


    data={  'keypoints0': list(kpsc_np),
            'keypoints1': list(kpst_np),
            'descrip0Scale': list(Snew_scale),
            'descrip1Scale': list(Gnew_scale),
            'descriptors0': list(Snew),
            'descriptors1': list(Gnew),
            'sc_location': (sc_location),
            'st_location': (st_location),
        }

    config = {
        'ma': {
            'descriptor_dim': Snew_scale.shape[0],
        }
    }
    mapper = mo.Matchsc_st(config.get('ma', {}))
    mapper.to(device)

    #optimizer = torch.optim.Adam(mapper.parameters(), lr=learning_rate)
    optimizer = torch.optim.Adamax(mapper.parameters(), lr=learning_rate)

    #scheduler=torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.5, patience=20, verbose=False, threshold=1e-4, threshold_mode='rel', cooldown=0, min_lr=0, eps=1e-08)

    mean_loss = []
    print("Start training")
    for epoch in range(1, num_epochs+1):
        '''
        # adjust the optimizer
        if epoch % 200 == 0:
            learning_rate /= 2
            optimizer = torch.optim.Adam(mapper.parameters(), lr=learning_rate, betas=(0.5, 0.99))
        '''
        t = time.time()
        epoch_loss = 0
        mapper.train()
        for k in data:
            if k != 'sc_location' and k!='st_location' and k!='cost_matrix':
                #print(type(pred[k]))
                if type(data[k]) == torch.Tensor:
                    #print(pred[k].shape)
                    data[k] = Variable(data[k].to(device))
                else:
                    data[k] = Variable(torch.stack(data[k]).to(device))
        pred,F_o = mapper(data)

        Loss,score0,score1 = mapper.get_loss(pred,data)
        #scheduler.step(Loss)

        #print(Loss)
        #mean_loss.append(Loss.item())
        optimizer.zero_grad()

        Loss.backward()
        optimizer.step()
        t1=time.time()-t
        if epoch%100==0:
            print ('Epoch {}, Loss: {:.3f},Score0: {:.3f},Time: {:.3f}' 
                .format(epoch, Loss.item(),score0.item(),t1)) 
        #mean_loss = []

    with torch.no_grad():
        mapping_matrix = softmax(pred[0,:,:], dim=-1).cpu().numpy()
        F_out = torch.sigmoid(F_o).cpu().numpy()

    adata_map = sc.AnnData(
        X=mapping_matrix,
        obs=adata_sc[:, training_genes].obs.copy(),
        var=adata_sp[:, training_genes].obs.copy(),
    )
    adata_map.obs["F_out"] = F_out
    G_predicted = adata_map.X.T @ S
    cos_sims = []
    for v1, v2 in zip(G.T, G_predicted.T):
        norm_sq = np.linalg.norm(v1) * np.linalg.norm(v2)
        cos_sims.append((v1 @ v2) / norm_sq)

    df_cs = pd.DataFrame(cos_sims, training_genes, columns=["train_score"])
    df_cs = df_cs.sort_values(by="train_score", ascending=False)
    adata_map.uns["train_genes_df"] = df_cs

    # Annotate sparsity of each training genes
    annotate_gene_sparsity(adata_sc)
    annotate_gene_sparsity(adata_sp)
    adata_map.uns["train_genes_df"]["sparsity_sc"] = adata_sc[
        :, training_genes
    ].var.sparsity
    adata_map.uns["train_genes_df"]["sparsity_sp"] = adata_sp[
        :, training_genes
    ].var.sparsity
    adata_map.uns["train_genes_df"]["sparsity_diff"] = (
        adata_sp[:, training_genes].var.sparsity
        - adata_sc[:, training_genes].var.sparsity
    )

    return adata_map


from pprint import pprint
import json
import re
import random
import operator
import collections
from pathlib import Path
random.seed(30)


def get_positive_edges(edge_file_id, path):
    fp = open(path, 'r')
    positive_edges = []
    for line in fp:
        edge = re.split(' ', line[:-1])
        edge = [0, int(edge[0]), int(edge[1])]
        positive_edges.append(edge)
    fp.close()
    print('No. of positive edges in', edge_file_id, len(positive_edges))

    return positive_edges


def get_ent2id(edge_file_id, path):
    # Get node ids
    fp = open(path, 'r')
    ent2id = {}
    for line in fp:
        line = re.split('\t', line[:-1])
        ent2id[line[0]] = int(line[1])
    fp.close()
    print('No. of nodes in', edge_file_id, len(ent2id))

    return ent2id


def split_positive_edges(positive_edges):
    # split positive edges
    no_pos = len(positive_edges)
    no_train = int(no_pos*0.8)
    no_test = int(no_pos*0.1)
    random.shuffle(positive_edges)
    train_edges = positive_edges[:no_train]
    valid_edges = positive_edges[no_train:-no_test]
    test_edges = positive_edges[-no_test:]
    print('No. of edges in training: {}, valid: {}, test: {}'
          .format(len(train_edges), len(valid_edges), len(test_edges)))

    return train_edges, valid_edges, test_edges


def generate_negative_edges(no_neg, ent2id, positive_edges):
    # generate negative edges

    negative_edges = []
    cnt = 0
    while 1:
        node_1, node_2 = random.sample(range(len(ent2id)), 2)
        edge = [0, node_1, node_2]
        reverse_edge = [0, node_2, node_1]
        if edge not in negative_edges and edge not in positive_edges and\
                reverse_edge not in negative_edges and reverse_edge not in positive_edges:
            edge = [0, node_1, node_2, 0]
            negative_edges.append(edge)
            cnt += 1
        if cnt == no_neg:
            break

    return negative_edges


def dump_data(ent2id, output_dir):
    # dump data
    ent2id = sorted(ent2id.items(), key=operator.itemgetter(1))
    ent2id = collections.OrderedDict(ent2id)
    out = json.dump(ent2id, open(
        '{}/ent2id.txt'.format(output_dir), 'w'))

    # dump edges
    tasks = {'train': train_edges, 'valid': valid_edges, 'test': test_edges}
    for task in tasks:
        print(task)
    #     print(tasks[task])
        fout = open('{}/{}.txt'.format(output_dir, task), 'w')
        for ee in tasks[task]:
            ee = [str(itm) for itm in ee]
            if task != 'train' and len(ee) == 3:
                ee.append('1')
            out = ' '.join(ee)
            fout.write(out)
            fout.write('\n')
        fout.close()


def dump_rel2id(path):
    rel2id = {'rel-0': 0}
    out = json.dump(rel2id, open(path, 'w'))


if __name__ == '__main__':
    cyber_data_dir = 'data/cyber'
    # Filter edges
    for edge_file_id in range(11, 18):
        print(edge_file_id)
        edge_file_id = str(edge_file_id)
        # Get Input
        fp = open(
            '{}/raw/testbed-{}jun-aggr.txt.edges'.format(cyber_data_dir, edge_file_id), 'r')
        cnt = 0
        all_edges = []
        for line in fp:
            edge = re.split('\t', line[:-1])
            reverse_edge = [edge[1], edge[0]]
            if edge not in all_edges and reverse_edge not in all_edges:
                all_edges.append(edge)
            cnt += 1

            if cnt % 50e3 == 0:
                print(cnt)

        fp.close()
        print(cnt, len(all_edges))

        # Create output directory
        day_output_dir = f'{cyber_data_dir}/processed/cyber_{edge_file_id}'
        Path(day_output_dir).mkdir(parents=True, exist_ok=True)

        # Dump Output
        fout = open(
            '{}/{}_filtered_edges.txt'.format(day_output_dir, edge_file_id), 'w')
        for ii in range(len(all_edges)):
            out = ' '.join(all_edges[ii])
            fout.write(out+'\n')
        fout.close()

    # Split positive edges and generate negative edges
    for edge_file_id in range(11, 18):
        print(edge_file_id)
        # Create output directory
        day_output_dir = f'{cyber_data_dir}/processed/cyber_{edge_file_id}'
        edge_file_id = str(edge_file_id)

        positive_edges_path = f"{day_output_dir}/{edge_file_id}_filtered_edges.txt"
        positive_edges = get_positive_edges(edge_file_id, positive_edges_path)

        ent2id_input_path = f'{cyber_data_dir}/raw/testbed-{edge_file_id}jun-aggr.txt.ids.tsv'
        ent2id = get_ent2id(edge_file_id, ent2id_input_path)

        train_edges, valid_edges, test_edges = split_positive_edges(
            positive_edges)

        no_neg = len(valid_edges)+len(test_edges)
        negative_edges = generate_negative_edges(
            no_neg, ent2id, positive_edges)
        random.shuffle(negative_edges)
        valid_edges += negative_edges[:-len(test_edges)]
        test_edges += negative_edges[-len(test_edges):]
        print('No. of edges in training: {}, valid: {}, test: {}'
              .format(len(train_edges), len(valid_edges), len(test_edges)))

        dump_data(ent2id, day_output_dir)
        rel2id_path = f"{day_output_dir}/rel2id.txt"
        dump_rel2id(rel2id_path)



"""
    Acknowledgement:
    This code builds on top of the CompGCN (https://github.com/malllabiisc/CompGCN) and
    PyTorch Geometric's message passing implementation.
"""

import sys
import torch
from ..CompGCN.model.compgcn_conv import CompGCNConv


class GCNTransform(torch.nn.Module):
    """
    Transforms a list of subgraphs extracted from a heterogeneous graph to their
    vector representations through a multi-relational GCN transformation.
    """

    def __init__(self, emb_sz, num_gcn_layers, node_edge_composition_func):
        """
        :param emb_sz:         Dimension of node and relation embeddings. Must be same.
        :param num_gcn_layers: Number of GCN transforms/rounds of message propoagation
                               that will happen on the input in forward().
        :param node_edge_composition_func: 'no_rel'|'mult'|'sub'|'circ_conv'
        """
        super(GCNTransform, self).__init__()
        self.embedding_dims = emb_sz
        self.num_gcn_layers = num_gcn_layers
        self.__gcn_layers = []
        if torch.cuda.is_available():
            self.device = torch.device("cuda")
        else:
            self.device = torch.device("cpu")

        for _ in range(num_gcn_layers):
            self.__gcn_layers.append(
                CompGCNConv(emb_sz, emb_sz, node_edge_composition_func).to(self.device)
            )
            # self.__conv1 = CompGCNConv(embed_dim, embed_dim, act=torch.tanh, params=self.p)

        self.drop = torch.nn.Dropout(0.3)
        return

    def __get_node_counts(self, subgraph_list):
        """
        Returns a list containing the number of nodes in each graph in the subgraph_list.
        """
        # print('subgraph_list in __get_node_counts', subgraph_list)
        node_counts = []
        for subgraph in subgraph_list:
            node_set = set()
            for edge in subgraph:
                node_set.add(edge[0])
                node_set.add(edge[1])
            node_counts.append(len(node_set))
        return node_counts

    def forward_subgraph_list(self, subgraph_list, in_node_emb, in_rel_emb):
        """
        Transforms a list of subgraphs [batch_sz] to their GCN transformed embeddings.
        :param subgraph_list:       A list of subgraphs where each subgraph is a list
                                    of (src, dst, relation) triples.
        :param in_node_emb: [batch_sz, max_nodes, emb_dim] dimension FloatTensor. max_nodes
                            refers to the maximum number of nodes across all subgraphs in
                            the subgraph_list. in_node_emb[i][j] will return the embedding
                            for node j in batch i.  This indices are zero based/normalized
                            to 0-(max_nodes-1).
        :param in_rel_emb:  [batch_sz, max_rels, emb_dim] dimension FloatTensor. max_rels
                            refers to the maximum number of relations across all subgraphs
                            in the subgraph_list. in_rel_emb[i][j] will return the embedding
                            for relation j in batch i. This indices are zero based/normalized
                            to 0-(max_relations-1).
        :ret out_node_emb:  Transformed version of in_node_emb with identical shape.
        :ret out_rel_emb:   Transformed version of in_rel_emb with identical shape.
        """
        # Obtain the number of nodes in each subgraph
        # node_counts = self.__get_node_counts(subgraph_list)
        # max_node_count = max(node_counts)
        out_node_emb = in_node_emb.clone()
        out_relation_emb = in_rel_emb.clone()

        for i, edge_list in enumerate(subgraph_list):
            x = out_node_emb[i]
            r = out_relation_emb[i]  # torch.cat([in_rel_emb, -in_rel_emb], dim=0)
            # Each subgraph in the subgraph_list is a list of (src, dst, relation)
            # MAJOR ASSUMPTION: the src/dst and relation ids are normalized such
            # that we can look up node_embeddings and relation_embeddings to fetch
            # corresponding embedding.

            # edge_index = [] # [(e[0], e[1]) for e in edge_list]
            # edge_types = [] # [e[2] for e in edge_list]
            # for e in edge_list:
            #     edge_index.append((e[0], e[1]))
            #     edge_types.append(e[2])
            # for e in edge_list:
            #     edge_index.append((e[1], e[0]))
            #     edge_types.append(e[2])

            # edge_index = torch.LongTensor(edge_index).t()
            # edge_types = torch.LongTensor(edge_types)

            # for i in range(self.num_gcn_layers):
            #     x, r = self.__gcn_layers[i](x, edge_index, edge_types, rel_embed=r)

            # x = self.drop(x)

            x, r = self.forward_subgraph(edge_list, x, r)
            out_node_emb[i] = x
            out_relation_emb[i] = r

        return out_node_emb, out_relation_emb

    def forward_subgraph(self, subgraph, in_node_emb, in_rel_emb):
        """
        Transforms a  subgraph's node and relation embeddings via GCN/Multi-Relational GCN.
        :param subgraph:    A list of (src, dst, relation) triples. Node ids are zero based
                            and normalized to range (0, number_of_nodes(subgraph)-1).
        :param in_node_emb: [N, emb_dim] dimension FloatTensor where N is at least number
                            of nodes in subgraph. in_node_emb[i] returns embedding for node i.
        :param in_rel_emb:  [N_e, emb_dim] dimension FloatTensor where N_e is the number of
                            distinct node types in subgraph/input graph. in_rel_emb[i] will
                            return the embedding for relation type i.
        :ret out_node_emb:  Transformed version of in_node_emb with identical shape.
        :ret out_rel_emb:   Transformed version of in_rel_emb with identical shape.
        """
        # Obtain the number of nodes in each subgraph
        # node_counts = self.__get_node_counts([subgraph])
        # max_node_count = max(node_counts)
        # out_node_emb = in_node_emb.clone()
        # out_relation_emb = in_rel_emb.clone()

        # edge_list = subgraph_list[i]
        edge_list = subgraph
        # Each subgraph in the subgraph_list is a list of (src, dst, relation)
        # MAJOR ASSUMPTION: the src/dst and relation ids are normalized such
        # that we can look up node_embeddings and relation_embeddings to fetch
        # corresponding embedding.
        edge_index = []  # [(e[0], e[1]) for e in edge_list]
        edge_types = []  # [e[2] for e in edge_list]
        for e in edge_list:
            edge_index.append((e[0], e[1]))
            edge_types.append(e[2])
        for e in edge_list:
            edge_index.append((e[1], e[0]))
            edge_types.append(e[2])

        edge_index = torch.LongTensor(edge_index).to(self.device).t()
        edge_types = torch.LongTensor(edge_types).to(self.device)

        x = in_node_emb.clone().to(self.device)
        r = in_rel_emb.clone().to(self.device)

        for i in range(self.num_gcn_layers):
            x, r = self.__gcn_layers[i](x, edge_index, edge_types, rel_embed=r)

        x = self.drop(x)
        return x, r

    def forward(self, option, subgraph_list, in_node_emb, in_rel_emb):
        if option == "subgraph_list":
            return self.forward_subgraph_list(subgraph_list, in_node_emb, in_rel_emb)
        elif option == "subgraph":
            return self.forward_subgraph(subgraph_list, in_node_emb, in_rel_emb)
        else:
            print("Unknown option: %s" % option)
            sys.exit(1)



from .helper import *
from .data_loader import *

# sys.path.append('./')
from .model.models import *
from orderedset import OrderedSet
class Runner(object):

	def load_data(self):
		"""
		Reading in raw triples and converts it into a standard format. 

		Parameters
		----------
		self.p.dataset:         Takes in the name of the dataset (FB15k-237)
		
		Returns
		-------
		self.ent2id:            Entity to unique identifier mapping
		self.id2rel:            Inverse mapping of self.ent2id
		self.rel2id:            Relation to unique identifier mapping
		self.num_ent:           Number of entities in the Knowledge graph
		self.num_rel:           Number of relations in the Knowledge graph
		self.embed_dim:         Embedding dimension used
		self.data['train']:     Stores the triples corresponding to training dataset
		self.data['valid']:     Stores the triples corresponding to validation dataset
		self.data['test']:      Stores the triples corresponding to test dataset
		self.data_iter:		The dataloader for different data splits

		"""

		ent_set, rel_set = OrderedSet(), OrderedSet()
		for split in ['train', 'test', 'valid']:
			for line in open('./data/{}/{}.txt'.format(self.p.dataset, split)):
				sub, rel, obj = map(str.lower, line.strip().split('\t'))
				ent_set.add(sub)
				rel_set.add(rel)
				ent_set.add(obj)

		self.ent2id = {ent: idx for idx, ent in enumerate(ent_set)}
		self.rel2id = {rel: idx for idx, rel in enumerate(rel_set)}
		self.rel2id.update({rel+'_reverse': idx+len(self.rel2id) for idx, rel in enumerate(rel_set)})

		self.id2ent = {idx: ent for ent, idx in self.ent2id.items()}
		self.id2rel = {idx: rel for rel, idx in self.rel2id.items()}

		self.p.num_ent		= len(self.ent2id)
		self.p.num_rel		= len(self.rel2id) // 2
		self.p.embed_dim	= self.p.k_w * self.p.k_h if self.p.embed_dim is None else self.p.embed_dim

		self.data = ddict(list)
		sr2o = ddict(set)

		for split in ['train', 'test', 'valid']:
			for line in open('./data/{}/{}.txt'.format(self.p.dataset, split)):
				sub, rel, obj = map(str.lower, line.strip().split('\t'))
				sub, rel, obj = self.ent2id[sub], self.rel2id[rel], self.ent2id[obj]
				self.data[split].append((sub, rel, obj))

				if split == 'train': 
					sr2o[(sub, rel)].add(obj)
					sr2o[(obj, rel+self.p.num_rel)].add(sub)

		self.data = dict(self.data)

		self.sr2o = {k: list(v) for k, v in sr2o.items()}
		for split in ['test', 'valid']:
			for sub, rel, obj in self.data[split]:
				sr2o[(sub, rel)].add(obj)
				sr2o[(obj, rel+self.p.num_rel)].add(sub)

		self.sr2o_all = {k: list(v) for k, v in sr2o.items()}
		self.triples  = ddict(list)

		for (sub, rel), obj in self.sr2o.items():
			self.triples['train'].append({'triple':(sub, rel, -1), 'label': self.sr2o[(sub, rel)], 'sub_samp': 1})

		for split in ['test', 'valid']:
			for sub, rel, obj in self.data[split]:
				rel_inv = rel + self.p.num_rel
				self.triples['{}_{}'.format(split, 'tail')].append({'triple': (sub, rel, obj), 	   'label': self.sr2o_all[(sub, rel)]})
				self.triples['{}_{}'.format(split, 'head')].append({'triple': (obj, rel_inv, sub), 'label': self.sr2o_all[(obj, rel_inv)]})

		self.triples = dict(self.triples)

		def get_data_loader(dataset_class, split, batch_size, shuffle=True):
			return  DataLoader(
					dataset_class(self.triples[split], self.p),
					batch_size      = batch_size,
					shuffle         = shuffle,
					num_workers     = max(0, self.p.num_workers),
					collate_fn      = dataset_class.collate_fn
				)

		self.data_iter = {
			'train':    	get_data_loader(TrainDataset, 'train', 	    self.p.batch_size),
			'valid_head':   get_data_loader(TestDataset,  'valid_head', self.p.batch_size),
			'valid_tail':   get_data_loader(TestDataset,  'valid_tail', self.p.batch_size),
			'test_head':   	get_data_loader(TestDataset,  'test_head',  self.p.batch_size),
			'test_tail':   	get_data_loader(TestDataset,  'test_tail',  self.p.batch_size),
		}

		self.edge_index, self.edge_type = self.construct_adj()

	def construct_adj(self):
		"""
		Constructor of the runner class

		Parameters
		----------
		
		Returns
		-------
		Constructs the adjacency matrix for GCN
		
		"""
		edge_index, edge_type = [], []

		for sub, rel, obj in self.data['train']:
			edge_index.append((sub, obj))
			edge_type.append(rel)

		# Adding inverse edges
		for sub, rel, obj in self.data['train']:
			edge_index.append((obj, sub))
			edge_type.append(rel + self.p.num_rel)

		edge_index	= torch.LongTensor(edge_index).to(self.device).t()
		edge_type	= torch.LongTensor(edge_type). to(self.device)

		return edge_index, edge_type

	def __init__(self, params):
		"""
		Constructor of the runner class

		Parameters
		----------
		params:         List of hyper-parameters of the model
		
		Returns
		-------
		Creates computational graph and optimizer
		
		"""
		self.p			= params
		self.logger		= get_logger(self.p.name, self.p.log_dir, self.p.config_dir)

		self.logger.info(vars(self.p))
		pprint(vars(self.p))

		if self.p.gpu != '-1' and torch.cuda.is_available():
			self.device = torch.device('cuda')
			torch.cuda.set_rng_state(torch.cuda.get_rng_state())
			torch.backends.cudnn.deterministic = True
		else:
			self.device = torch.device('cpu')

		self.load_data()
		self.model        = self.add_model(self.p.model, self.p.score_func)
		self.optimizer    = self.add_optimizer(self.model.parameters())


	def add_model(self, model, score_func):
		"""
		Creates the computational graph

		Parameters
		----------
		model_name:     Contains the model name to be created
		
		Returns
		-------
		Creates the computational graph for model and initializes it
		
		"""
		model_name = '{}_{}'.format(model, score_func)

		if   model_name.lower()	== 'compgcn_transe': 	model = CompGCN_TransE(self.edge_index, self.edge_type, params=self.p)
		elif model_name.lower()	== 'compgcn_distmult': 	model = CompGCN_DistMult(self.edge_index, self.edge_type, params=self.p)
		elif model_name.lower()	== 'compgcn_conve': 	model = CompGCN_ConvE(self.edge_index, self.edge_type, params=self.p)
		else: raise NotImplementedError

		model.to(self.device)
		return model

	def add_optimizer(self, parameters):
		"""
		Creates an optimizer for training the parameters

		Parameters
		----------
		parameters:         The parameters of the model
		
		Returns
		-------
		Returns an optimizer for learning the parameters of the model
		
		"""
		return torch.optim.Adam(parameters, lr=self.p.lr, weight_decay=self.p.l2)

	def read_batch(self, batch, split):
		"""
		Function to read a batch of data and move the tensors in batch to CPU/GPU

		Parameters
		----------
		batch: 		the batch to process
		split: (string) If split == 'train', 'valid' or 'test' split

		
		Returns
		-------
		Head, Relation, Tails, labels
		"""
		if split == 'train':
			triple, label = [ _.to(self.device) for _ in batch]
			return triple[:, 0], triple[:, 1], triple[:, 2], label
		else:
			triple, label = [ _.to(self.device) for _ in batch]
			return triple[:, 0], triple[:, 1], triple[:, 2], label

	def save_model(self, save_path):
		"""
		Function to save a model. It saves the model parameters, best validation scores,
		best epoch corresponding to best validation, state of the optimizer and all arguments for the run.

		Parameters
		----------
		save_path: path where the model is saved
		
		Returns
		-------
		"""
		state = {
			'state_dict'	: self.model.state_dict(),
			'best_val'	: self.best_val,
			'best_epoch'	: self.best_epoch,
			'optimizer'	: self.optimizer.state_dict(),
			'args'		: vars(self.p)
		}
		torch.save(state, save_path)

	def load_model(self, load_path):
		"""
		Function to load a saved model

		Parameters
		----------
		load_path: path to the saved model
		
		Returns
		-------
		"""
		state			= torch.load(load_path)
		state_dict		= state['state_dict']
		self.best_val		= state['best_val']
		self.best_val_mrr	= self.best_val['mrr'] 

		self.model.load_state_dict(state_dict)
		self.optimizer.load_state_dict(state['optimizer'])

	def evaluate(self, split, epoch):
		"""
		Function to evaluate the model on validation or test set

		Parameters
		----------
		split: (string) If split == 'valid' then evaluate on the validation set, else the test set
		epoch: (int) Current epoch count
		
		Returns
		-------
		resutls:			The evaluation results containing the following:
			results['mr']:         	Average of ranks_left and ranks_right
			results['mrr']:         Mean Reciprocal Rank
			results['hits@k']:      Probability of getting the correct preodiction in top-k ranks based on predicted score

		"""
		left_results  = self.predict(split=split, mode='tail_batch')
		right_results = self.predict(split=split, mode='head_batch')
		results       = get_combined_results(left_results, right_results)
		self.logger.info('[Epoch {} {}]: MRR: Tail : {:.5}, Head : {:.5}, Avg : {:.5}'.format(epoch, split, results['left_mrr'], results['right_mrr'], results['mrr']))
		return results

	def predict(self, split='valid', mode='tail_batch'):
		"""
		Function to run model evaluation for a given mode

		Parameters
		----------
		split: (string) 	If split == 'valid' then evaluate on the validation set, else the test set
		mode: (string):		Can be 'head_batch' or 'tail_batch'
		
		Returns
		-------
		resutls:			The evaluation results containing the following:
			results['mr']:         	Average of ranks_left and ranks_right
			results['mrr']:         Mean Reciprocal Rank
			results['hits@k']:      Probability of getting the correct preodiction in top-k ranks based on predicted score

		"""
		self.model.eval()

		with torch.no_grad():
			results = {}
			train_iter = iter(self.data_iter['{}_{}'.format(split, mode.split('_')[0])])

			for step, batch in enumerate(train_iter):
				sub, rel, obj, label	= self.read_batch(batch, split)
				pred			= self.model.forward(sub, rel)
				b_range			= torch.arange(pred.size()[0], device=self.device)
				target_pred		= pred[b_range, obj]
				pred 			= torch.where(label.byte(), -torch.ones_like(pred) * 10000000, pred)
				pred[b_range, obj] 	= target_pred
				ranks			= 1 + torch.argsort(torch.argsort(pred, dim=1, descending=True), dim=1, descending=False)[b_range, obj]

				ranks 			= ranks.float()
				results['count']	= torch.numel(ranks) 		+ results.get('count', 0.0)
				results['mr']		= torch.sum(ranks).item() 	+ results.get('mr',    0.0)
				results['mrr']		= torch.sum(1.0/ranks).item()   + results.get('mrr',   0.0)
				for k in range(10):
					results['hits@{}'.format(k+1)] = torch.numel(ranks[ranks <= (k+1)]) + results.get('hits@{}'.format(k+1), 0.0)

				if step % 100 == 0:
					self.logger.info('[{}, {} Step {}]\t{}'.format(split.title(), mode.title(), step, self.p.name))

		return results


	def run_epoch(self, epoch, val_mrr = 0):
		"""
		Function to run one epoch of training

		Parameters
		----------
		epoch: current epoch count
		
		Returns
		-------
		loss: The loss value after the completion of one epoch
		"""
		self.model.train()
		losses = []
		train_iter = iter(self.data_iter['train'])

		for step, batch in enumerate(train_iter):
			self.optimizer.zero_grad()
			sub, rel, obj, label = self.read_batch(batch, 'train')

			pred	= self.model.forward(sub, rel)
			loss	= self.model.loss(pred, label)

			loss.backward()
			self.optimizer.step()
			losses.append(loss.item())

			if step % 100 == 0:
				self.logger.info('[E:{}| {}]: Train Loss:{:.5},  Val MRR:{:.5}\t{}'.format(epoch, step, np.mean(losses), self.best_val_mrr, self.p.name))

		loss = np.mean(losses)
		self.logger.info('[Epoch:{}]:  Training Loss:{:.4}\n'.format(epoch, loss))
		return loss


	def fit(self):
		"""
		Function to run training and evaluation of model

		Parameters
		----------
		
		Returns
		-------
		"""
		self.best_val_mrr, self.best_val, self.best_epoch, val_mrr = 0., {}, 0, 0.
		save_path = os.path.join('./checkpoints', self.p.name)

		if self.p.restore:
			self.load_model(save_path)
			self.logger.info('Successfully Loaded previous model')

		for epoch in range(self.p.max_epochs):
			train_loss  = self.run_epoch(epoch, val_mrr)
			val_results = self.evaluate('valid', epoch)

			if val_results['mrr'] > self.best_val_mrr:
				self.best_val	   = val_results
				self.best_val_mrr  = val_results['mrr']
				self.best_epoch	   = epoch
				self.save_model(save_path)

			self.logger.info('[Epoch {}]: Training Loss: {:.5}, Valid MRR: {:.5}\n\n'.format(epoch, train_loss, self.best_val_mrr))

		self.logger.info('Loading best model, Evaluating on Test data')
		self.load_model(save_path)
		test_results = self.evaluate('test', epoch)

if __name__ == '__main__':
	parser = argparse.ArgumentParser(description='Parser For Arguments', formatter_class=argparse.ArgumentDefaultsHelpFormatter)

	parser.add_argument('-name',		default='testrun',					help='Set run name for saving/restoring models')
	parser.add_argument('-data',		dest='dataset',         default='FB15k-237',            help='Dataset to use, default: FB15k-237')
	parser.add_argument('-model',		dest='model',		default='compgcn',		help='Model Name')
	parser.add_argument('-score_func',	dest='score_func',	default='conve',		help='Score Function for Link prediction')
	parser.add_argument('-opn',             dest='opn',             default='corr',                 help='Composition Operation to be used in CompGCN')

	parser.add_argument('-batch',           dest='batch_size',      default=128,    type=int,       help='Batch size')
	parser.add_argument('-gamma',		type=float,             default=40.0,			help='Margin')
	parser.add_argument('-gpu',		type=str,               default='0',			help='Set GPU Ids : Eg: For CPU = -1, For Single GPU = 0')
	parser.add_argument('-epoch',		dest='max_epochs', 	type=int,       default=500,  	help='Number of epochs')
	parser.add_argument('-l2',		type=float,             default=0.0,			help='L2 Regularization for Optimizer')
	parser.add_argument('-lr',		type=float,             default=0.001,			help='Starting Learning Rate')
	parser.add_argument('-lbl_smooth',      dest='lbl_smooth',	type=float,     default=0.1,	help='Label Smoothing')
	parser.add_argument('-num_workers',	type=int,               default=10,                     help='Number of processes to construct batches')
	parser.add_argument('-seed',            dest='seed',            default=41504,  type=int,     	help='Seed for randomization')

	parser.add_argument('-restore',         dest='restore',         action='store_true',            help='Restore from the previously saved model')
	parser.add_argument('-bias',            dest='bias',            action='store_true',            help='Whether to use bias in the model')

	parser.add_argument('-num_bases',	dest='num_bases', 	default=-1,   	type=int, 	help='Number of basis relation vectors to use')
	parser.add_argument('-init_dim',	dest='init_dim',	default=100,	type=int,	help='Initial dimension size for entities and relations')
	parser.add_argument('-gcn_dim',	  	dest='gcn_dim', 	default=200,   	type=int, 	help='Number of hidden units in GCN')
	parser.add_argument('-embed_dim',	dest='embed_dim', 	default=None,   type=int, 	help='Embedding dimension to give as input to score function')
	parser.add_argument('-gcn_layer',	dest='gcn_layer', 	default=1,   	type=int, 	help='Number of GCN Layers to use')
	parser.add_argument('-gcn_drop',	dest='dropout', 	default=0.1,  	type=float,	help='Dropout to use in GCN Layer')
	parser.add_argument('-hid_drop',  	dest='hid_drop', 	default=0.3,  	type=float,	help='Dropout after GCN')

	# ConvE specific hyperparameters
	parser.add_argument('-hid_drop2',  	dest='hid_drop2', 	default=0.3,  	type=float,	help='ConvE: Hidden dropout')
	parser.add_argument('-feat_drop', 	dest='feat_drop', 	default=0.3,  	type=float,	help='ConvE: Feature Dropout')
	parser.add_argument('-k_w',	  	dest='k_w', 		default=10,   	type=int, 	help='ConvE: k_w')
	parser.add_argument('-k_h',	  	dest='k_h', 		default=20,   	type=int, 	help='ConvE: k_h')
	parser.add_argument('-num_filt',  	dest='num_filt', 	default=200,   	type=int, 	help='ConvE: Number of filters in convolution')
	parser.add_argument('-ker_sz',    	dest='ker_sz', 		default=7,   	type=int, 	help='ConvE: Kernel size to use')

	parser.add_argument('-logdir',          dest='log_dir',         default='./log/',               help='Log directory')
	parser.add_argument('-config',          dest='config_dir',      default='./config/',            help='Config directory')
	args = parser.parse_args()

	if not args.restore: args.name = args.name + '_' + time.strftime('%d_%m_%Y') + '_' + time.strftime('%H:%M:%S')

	set_gpu(args.gpu)
	np.random.seed(args.seed)
	torch.manual_seed(args.seed)

	model = Runner(args)
	model.fit()



import json
import os
import pickle
import shutil
import time

import numpy as np
import torch
import torch.nn

import torch.optim as optim
from torch.autograd import Variable

from .ccc_model.ccc_model import CCC, FinetuneLayer
from .ccc_model.process_walks_gcn import Processing_GCN_Walks
from .utils.context_metrics import PathMetrics
from .utils.data_utils import create_finetune_batches2
from .utils.utils import load_pickle, show_progress

class FocalLoss(torch.nn.Module):
    def __init__(self, alpha=1, gamma=2, reduction='mean'):
        super().__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.reduction = reduction

    def forward(self, inputs, targets):
        BCE_loss = torch.nn.BCEWithLogitsLoss(reduction='mean')(inputs, targets)
        pt = torch.exp(-BCE_loss)
        F_loss = self.alpha * (1 - pt) ** self.gamma * BCE_loss
        if self.reduction == 'mean':
            return torch.mean(F_loss)
        elif self.reduction == 'sum':
            return torch.sum(F_loss)
        else:
            return F_loss


def setup_finetuning_input(args, attr_graph, context_gen):
    # finetune
    #print("\n Generate data for finetuning ...")
    num_batches = {}
    finetune_path = args.outdir + "/finetune/"
    if not os.path.exists(finetune_path):
        os.mkdir(finetune_path)

    if os.path.exists(finetune_path + "/finetune_walks.txt"):
        finetune_walks_per_task = json.load(open(finetune_path + "/finetune_walks.txt"))
    else:

        (
            train_edges,
            valid_edges,
            test_edges,
        ) = attr_graph.generate_link_prediction_dataset(
            finetune_path, fr_valid_edges=0.1, fr_test_edges=0.1
        )
        finetune_walks_per_task = context_gen.get_finetune_subgraphs(
            train_edges,
            valid_edges,
            test_edges,
            args.beam_width,
            args.max_length,
            args.path_option,
        )
        json.dump(
            finetune_walks_per_task, open(finetune_path + "/finetune_walks.txt", "w")
        )

        #print(
        #    "Number of train, valid, test",
        #    len(train_edges),
        #    len(valid_edges),
        #    len(test_edges),
        #)
    num_batches = create_finetune_batches2(
        finetune_walks_per_task, finetune_path, args.data_name, args.ft_batch_size
    )
    #print("No. of batches for finetuning:", num_batches)

    return num_batches


def run_finetuning_wkfl2(
    args, attr_graph, no_batches, pretrained_node_embedding, ent2id, rel2id, epoch=-1
):

    # data_path = args.data_path + args.data_name + "/"
    outdir = args.outdir + "/"
    finetune_path = outdir + "finetune/"
    ft_out_dir = finetune_path + "results"
    try:
        shutil.rmtree(ft_out_dir)
        os.mkdir(ft_out_dir)
    except:  # FIXME - need to replace bare except
        os.mkdir(ft_out_dir)
    relations = attr_graph.relation_to_id
    nodeid2rowid = attr_graph.get_nodeid2rowid()
    walk_processor = Processing_GCN_Walks(
        nodeid2rowid, relations, args.n_pred, args.max_length, args.max_pred
    )

    # process minibatch for finetune(ft)
    #print("\n processing minibaches for finetuning:")
    ft_batch_input_file = os.path.join(
        finetune_path, args.data_name + "_ft_batch_input.pickled"
    )
    if os.path.exists(ft_batch_input_file):
        #print("loading saved files ...")
        ft_batch_input = load_pickle(ft_batch_input_file)
    else:
        ft_batch_input = {}
        tasks = ["train", "valid", "test"]
        for task in tasks:
            #print(task)
            ft_batch_input[task] = {}
            for batch_id in range(no_batches[task]):
                (
                    subgraphs,
                    all_nodes,
                    labels,
                ) = walk_processor.process_finetune_minibatch(
                    finetune_path, args.data_name, task, batch_id
                )
                ft_batch_input[task][batch_id] = [subgraphs, all_nodes, labels]
        pickle.dump(ft_batch_input, open(ft_batch_input_file, "wb"))

    if args.is_pre_trained:
        pretrained_node_embedding_tensor = pretrained_node_embedding
    else:
        pretrained_node_embedding_tensor = None

    ccc = CCC(
        args.n_layers,
        args.device,
        args.d_model,
        args.d_k,
        args.d_v,
        args.d_ff,
        args.n_heads,
        attr_graph,
        pretrained_node_embedding_tensor,
        args.is_pre_trained,
        args.base_embedding_dim,
        args.max_length,
        args.num_gcn_layers,
        args.node_edge_composition_func,
        args.gcn_option,
        args.get_bert_encoder_embeddings,
        ent2id,
        rel2id,
    )

    #print("Begin finetuning")
    # load pre-trained model
    fbest = open(os.path.join(outdir, "best_epoch_id.txt"), "r")
    for line in fbest:
        tmp = line
    best_epoch = int(tmp[:-1])
    #print("best_epoch: ", best_epoch)
    fbest.close()
    if epoch == -1:
        fl_ = os.path.join(outdir, "bert_{}.model".format(best_epoch))
    else:
        fl_ = os.path.join(outdir, "bert_{}.model".format(epoch))
    #print("LOADING PRE_TRAIN model from ", fl_)
    ccc.load_state_dict(
        torch.load(fl_, map_location=lambda storage, loc: storage)
    )

    # run in fine tuning mode
    ccc.set_fine_tuning()
    # testing
    #print("Begin Testing")
    ccc.eval()

    pred_data = {}
    true_data = {}
    path_metrics_graphbert_pre = PathMetrics(
        args.device,
        None,
        attr_graph.G,
        ent2id,
        args.d_model,
        args.ft_d_ff,
        args.ft_layer,
        args.ft_drop_rate,
        attr_graph,
        args.ft_input_option,
        args.n_layers,
    )
    with torch.no_grad():
        for data_type in ["valid", "test"]:
            pred_data[data_type] = []
            true_data[data_type] = []
            for batch_id in range(no_batches[data_type]):
                if batch_id % 100 == 0:
                    print("Evaluating {} batch {}".format(data_type, batch_id))
                subgraphs, all_nodes, labels = ft_batch_input[data_type][batch_id]

                masked_pos = torch.randn(args.batch_size, 2)
                masked_nodes = Variable(
                    torch.LongTensor([[] for ii in range(args.ft_batch_size)])
                )
                output, _, _ = ccc(
                    subgraphs, all_nodes, masked_pos, masked_nodes
                )
                source_embed = output[:, 0, :].unsqueeze(1)
                target_embed = output[:, 1, :].unsqueeze(1).transpose(1, 2)
                score = torch.bmm(source_embed, target_embed).squeeze(1)
                #score = torch.sigmoid(score).data.cpu().numpy().tolist()
                labels = labels.data.cpu().numpy().tolist()
                if data_type == "test":
                    path_metrics_graphbert_pre.update_batch_graphbert(
                        subgraphs, score, labels, output, all_nodes, "graphbert_pre"
                    )

                for ii, _ in enumerate(score):
                    pred_data[data_type].append(score[ii][0])
                    true_data[data_type].append(labels[ii][0])
                # print(len(score), len(labels))

        path_dict = path_metrics_graphbert_pre.finalize()
        '''
        json.dump(
            path_dict, open(outdir + "path_metrics_graphbert_pre.json", "w"), indent=4
        )
        '''

    return pred_data, true_data


def run_finetuning_wkfl3(
    args, attr_graph, no_batches, pretrained_node_embedding, ent2id, rel2id
):

    # data_path = args.data_path + args.data_name + "/"
    outdir = args.outdir + "/"
    finetune_path = outdir + "finetune/"
    ft_out_dir = finetune_path + "results"
    try:
        shutil.rmtree(ft_out_dir)
        os.mkdir(ft_out_dir)
    except:  # FIXME - need to replace bare except
        os.mkdir(ft_out_dir)
    relations = attr_graph.relation_to_id
    nodeid2rowid = attr_graph.get_nodeid2rowid()
    walk_processor = Processing_GCN_Walks(
        nodeid2rowid, relations, args.n_pred, args.max_length, args.max_pred
    )

    # process minibatch for finetune(ft)
    #print("\n processing minibaches for finetuning:")
    ft_batch_input_file = os.path.join(
        finetune_path, args.data_name + "_ft_batch_input.pickled"
    )
    if os.path.exists(ft_batch_input_file):
        #print("loading saved files ...")
        ft_batch_input = load_pickle(ft_batch_input_file)
    else:
        ft_batch_input = {}
        tasks = ["train", "valid", "test"]
        for task in tasks:
            #print(task)
            ft_batch_input[task] = {}
            for batch_id in range(no_batches[task]):
                (
                    subgraphs,
                    all_nodes,
                    labels,
                ) = walk_processor.process_finetune_minibatch(
                    finetune_path, args.data_name, task, batch_id
                )
                ft_batch_input[task][batch_id] = [subgraphs, all_nodes, labels]
        pickle.dump(ft_batch_input, open(ft_batch_input_file, "wb"))

    if args.is_pre_trained:
        pretrained_node_embedding_tensor = pretrained_node_embedding
    else:
        pretrained_node_embedding_tensor = None

    ccc = CCC(
        args.n_layers,
        args.device,
        args.d_model,
        args.d_k,
        args.d_v,
        args.d_ff,
        args.n_heads,
        attr_graph,
        pretrained_node_embedding_tensor,
        args.is_pre_trained,
        args.base_embedding_dim,
        args.max_length,
        args.num_gcn_layers,
        args.node_edge_composition_func,
        args.gcn_option,
        args.get_bert_encoder_embeddings,
        ent2id,
        rel2id,
    )

    pretrained_node_embedding_tensor = ccc.gcn_graph_encoder.node_embedding

    ft_linear = FinetuneLayer(
        args.device,
        args.d_model,
        args.ft_d_ff,
        args.ft_layer,
        args.ft_drop_rate,
        attr_graph,
        args.ft_input_option,
        args.n_layers,
    )
    criterion = torch.nn.BCEWithLogitsLoss(reduction='mean')
    #criterion = FocalLoss(reduction="mean")
    #criterion = torch.nn.BCELoss(reduction="mean")
    optimizer = optim.Adam(ft_linear.parameters(), args.ft_lr,weight_decay=0.0001)

    #print("Begin finetuning")
    # load pre-trained model
    fbest = open(os.path.join(outdir, "best_epoch_id.txt"), "r")
    for line in fbest:
        tmp = line
    best_epoch = int(tmp[:-1])
    #print("best_epoch in pretraining: ", best_epoch)
    fbest.close()
    fl_ = os.path.join(outdir, "bert_{}.model".format(best_epoch))
    #print("LOADING PRE_TRAIN model from ", fl_)
    ccc.load_state_dict(
        torch.load(fl_, map_location=lambda storage, loc: storage)
    )

    # run in fine tuning mode
    ccc.set_fine_tuning()

    print("\n Begin Training")
    loss_dev_final = []
    for epoch in range(args.ft_n_epochs):
        loss_arr = []
        loss_dev_min = 1e6
        start_time = time.time()
        for batch_id in range(no_batches["train"]):
            subgraphs, all_nodes, labels = walk_processor.process_finetune_minibatch(
                finetune_path, args.data_name, "train", batch_id
            )

            with torch.no_grad():
                # random mask
                masked_pos = torch.randn(args.batch_size, 2)
                masked_nodes = Variable(
                    torch.LongTensor([[] for ii in range(args.ft_batch_size)])
                )
                _, layer_output, _ = ccc(
                    subgraphs, all_nodes, masked_pos, masked_nodes
                )
            pred_scores, _, _ = ft_linear(layer_output)
            # print(" Size of ypred, ytrue", pred_scores.size(), labels.size())
            loss = criterion(pred_scores.squeeze(1), labels.squeeze(1).to(args.device))

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            loss_arr.append(loss.data.cpu().numpy().tolist())

            ccnt = batch_id % args.ft_checkpoint
            if batch_id > 0 and batch_id % args.ft_checkpoint == 0:
                ccnt = args.ft_checkpoint
            '''
            message = "Loss: {}, AvgLoss: {}{}".format(
                np.around(loss.data.cpu().numpy().tolist(), 4),
                np.around(np.average(loss_arr).tolist(), 4),
                " " * 10,
            )
            
            show_progress(
                ccnt, min(args.ft_checkpoint, no_batches["train"]), message
            )
            '''
            if (
                batch_id % args.ft_checkpoint == 0 and batch_id > 0
            ) or batch_id == no_batches["train"] - 1:
                '''
                print(
                    "\nBatch: {}, Loss: {}, AvgLoss: {}".format(
                        batch_id,
                        np.around(loss.data.cpu().numpy().tolist(), 4),
                        np.around(np.average(loss_arr).tolist(), 4),
                    )
                )
                '''
                # validation
                ccc.eval()
                ft_linear.eval()
                pred_data_valid = []
                true_data_valid = []
                with torch.no_grad():
                    loss_dev_arr = []
                    for batch_dev_id in range(no_batches["valid"]):
                        subgraphs, all_nodes, labels = ft_batch_input["valid"][
                            batch_dev_id
                        ]

                        # random mask
                        masked_pos = torch.randn(args.batch_size, 2)
                        masked_nodes = Variable(
                            torch.LongTensor([[] for ii in range(args.ft_batch_size)])
                        )
                        _, layer_output, _ = ccc(
                            subgraphs, all_nodes, masked_pos, masked_nodes
                        )
                        pred_scores, _, _ = ft_linear(layer_output)
                        loss = criterion(
                            pred_scores.squeeze(1), labels.squeeze(1).to(args.device)
                        )
                        loss_dev_arr.append(loss.data.cpu().numpy().tolist())

                        score = pred_scores.data.cpu().numpy().tolist()
                        labels = labels.data.cpu().numpy().tolist()
                        for ii, _ in enumerate(score):
                            pred_data_valid.append(score[ii][0])
                            true_data_valid.append(labels[ii][0])

                    loss_dev_avg = np.average(loss_dev_arr)
                    print(
                        "MinLoss: {}, CurLoss: {}".format(
                            np.around(loss_dev_min, 4), np.around(loss_dev_avg, 4)
                        )
                    )

                    if loss_dev_avg < loss_dev_min:
                        loss_dev_min = loss_dev_avg
                        fmodel = open(
                            os.path.join(
                                ft_out_dir, "finetune_" + str(epoch) + ".model"
                            ),
                            "wb",
                        )
                        torch.save(ft_linear.state_dict(), fmodel)
                        fmodel.close()

                ccc.train()
                ft_linear.train()
        loss_dev_final.append(loss_dev_min)
        print("\nFinetune epoch: ", epoch)
        print("MinLoss: ", np.around(loss_dev_min, 4))
        end_time = time.time()
        print("epoch time: (s)", (end_time - start_time))
    best_epoch = np.argsort(loss_dev_final)[0]
    print("\nBest Epoch: {}".format(best_epoch))

    np.save(f"{args.outdir}/finetuning_loss.npy", loss_dev_final)
    # testing
    print("Begin Testing")
    ccc.eval()
    ft_linear.eval()

    fl_ = os.path.join(ft_out_dir, "finetune_{}.model".format(best_epoch))
    ft_linear.load_state_dict(
        torch.load(fl_, map_location=lambda storage, loc: storage)
    )

    pred_data = []
    true_data = []
    path_metrics_pretrained = PathMetrics(
        args.device,
        pretrained_node_embedding_tensor,
        attr_graph.G,
        ent2id,
        args.d_model,
        args.ft_d_ff,
        args.ft_layer,
        args.ft_drop_rate,
        attr_graph,
        args.ft_input_option,
        args.n_layers,
    )
    path_metrics_graphbert_ft = PathMetrics(
        args.device,
        None,
        attr_graph.G,
        ent2id,
        args.d_model,
        args.ft_d_ff,
        args.ft_layer,
        args.ft_drop_rate,
        attr_graph,
        args.ft_input_option,
        args.n_layers,
    )
    with torch.no_grad():
        att_weights = {}
        final_emd = {}
        for batch_test_id in range(no_batches["test"]):
            if batch_test_id % 100 == 0:
                print("Evaluating test batch {}".format(batch_test_id))
            subgraphs, all_nodes, labels = ft_batch_input["test"][batch_test_id]
            masked_pos = torch.randn(args.batch_size, 2)
            masked_nodes = Variable(
                torch.LongTensor([[] for ii in range(args.ft_batch_size)])
            )
            _, layer_output, att_output = ccc(
                subgraphs, all_nodes, masked_pos, masked_nodes
            )
            score, src_embedding, dst_embedding = ft_linear(layer_output)
            score = score.data.cpu().numpy().tolist()
            labels = labels.data.cpu().numpy().tolist()

            att_weights[batch_test_id] = {
                "attentions": att_output,
                "subgraphs": subgraphs,
                "nodes_sequence": all_nodes,
                "labels": labels,
            }
            final_emd[batch_test_id] = {
                "src_emd": src_embedding,
                "dst_emd": dst_embedding,
                "subgraphs": subgraphs,
                "nodes_sequence": all_nodes,
                "labels": labels,
            }
            path_metrics_pretrained.update_batch(subgraphs, labels)
            path_metrics_graphbert_ft.update_batch_graphbert(
                subgraphs, score, labels, layer_output, all_nodes, "graphbert_ft"
            )

            for ii, _ in enumerate(score):
                pred_data.append(score[ii][0])
                true_data.append(labels[ii][0])
            # print(len(score), len(labels))

        # dump path metrics
        path_dict = path_metrics_pretrained.finalize()
        json.dump(
            path_dict, open(outdir + "path_metrics_pretrained.json", "w"), indent=4
        )
        path_dict = path_metrics_graphbert_ft.finalize()
        json.dump(
            path_dict, open(outdir + "path_metrics_graphbert_ft.json", "w"), indent=4
        )
        # dump attention weights
        att_weights_file = outdir + "attention_weights.pickled"
        pickle.dump(att_weights, open(att_weights_file, "wb"))
        # dump embeddings for visulization#nodes sequences
        # flag = len(nodes_sequence[ii])
        # for jj in range(len(nodes_sequence[ii])):
        #     if nodes_sequence[ii][jj] == 10099:
        #         flag = jj
        #         break
        # nodes_sequence[ii] = nodes_sequence[ii][:flag]
        final_emd_file = outdir + "final_emd_test.pickled"
        pickle.dump(final_emd, open(final_emd_file, "wb"))

    return pred_data, true_data, pred_data_valid, true_data_valid



import math

import numpy as np
import torch

from torch.autograd import Variable

from ..encoding.gcn_graph_encoder import GCNGraphEncoder
from ..encoding.gcn_transform import GCNTransform


def get_attn_pad_mask(seq_q, seq_k, pad_id,device):
    batch_size, len_q = len(seq_q), len(seq_q[0])
    batch_size, len_k = len(seq_k), len(seq_k[0])
    # print(batch_size, len_q, len_k)
    pad_attn_mask = []
    for itm in seq_k:
        tmp_mask = []
        for sub in itm:
            if sub == pad_id:
                tmp_mask.append(True)
            else:
                tmp_mask.append(False)
        pad_attn_mask.append(tmp_mask)
        # print(tmp_mask)
    # print('mask', len(pad_attn_mask), len(pad_attn_mask[0]))
    pad_attn_mask = Variable(torch.ByteTensor(pad_attn_mask)).unsqueeze(1)
    pad_attn_mask = pad_attn_mask.to(device)

    return pad_attn_mask.expand(batch_size, len_q, len_k)  # batch_size x len_q x len_k


def gelu(x):
    """"Implementation of the gelu activation function by Hugging Face."""
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))


class ScaledDotProductAttention(torch.nn.Module):
    def __init__(self, d_k):

        super(ScaledDotProductAttention, self).__init__()
        self.d_k = d_k

    def forward(self, Q, K, V, attn_mask=None):
        # print('mask', attn_mask.size())
        scores = torch.matmul(Q, K.transpose(-1, -2)) / np.sqrt(self.d_k)
        scores.masked_fill_(attn_mask == True, -1e9)
        attn = torch.nn.Softmax(dim=-1)(scores)
        context = torch.matmul(attn, V)

        return context, attn


class MultiHeadAttention(torch.nn.Module):
    def __init__(self, d_model, d_k, d_v, n_heads):

        super(MultiHeadAttention, self).__init__()
        self.n_heads = n_heads
        self.d_k = d_k
        self.d_v = d_v
        self.d_model = d_model

        self.W_Q = torch.nn.Linear(d_model, d_k * n_heads)
        self.W_K = torch.nn.Linear(d_model, d_k * n_heads)
        self.W_V = torch.nn.Linear(d_model, d_v * n_heads)
        self.scaled_dot_prod_attn = ScaledDotProductAttention(d_k)
        self.wrap = torch.nn.Linear(self.n_heads * self.d_v, self.d_model)
        self.layerNorm = torch.nn.LayerNorm(self.d_model)

    def forward(self, Q, K, V, attn_mask=None):
        residual, batch_size = Q, Q.size(0)
        q_s = self.W_Q(Q).view(batch_size, -1, self.n_heads, self.d_k).transpose(1, 2)
        k_s = self.W_K(K).view(batch_size, -1, self.n_heads, self.d_k).transpose(1, 2)
        v_s = self.W_V(V).view(batch_size, -1, self.n_heads, self.d_v).transpose(1, 2)

        if attn_mask is not None:
            attn_mask = attn_mask.unsqueeze(1).repeat(1, self.n_heads, 1, 1)
        context, attn = self.scaled_dot_prod_attn(q_s, k_s, v_s, attn_mask=attn_mask)
        context = (
            context.transpose(1, 2)
            .contiguous()
            .view(batch_size, -1, self.n_heads * self.d_v)
        )
        output = self.wrap(context)

        return self.layerNorm(output + residual), attn


class PoswiseFeedForwardNet(torch.nn.Module):
    def __init__(self, d_model, d_ff):

        super(PoswiseFeedForwardNet, self).__init__()
        self.fc1 = torch.nn.Linear(d_model, d_ff)
        self.fc2 = torch.nn.Linear(d_ff, d_model)

    def forward(self, x):

        return self.fc2(gelu(self.fc1(x)))


class EncoderLayer(torch.nn.Module):
    def __init__(self, d_model, d_k, d_v, d_ff, n_heads):

        super(EncoderLayer, self).__init__()
        self.enc_self_attn = MultiHeadAttention(d_model, d_k, d_v, n_heads)
        self.pos_ffn = PoswiseFeedForwardNet(d_model, d_ff)

    def forward(self, enc_inputs, enc_self_attn_mask):
        enc_outputs, attn = self.enc_self_attn(
            enc_inputs, enc_inputs, enc_inputs, enc_self_attn_mask
        )  # enc_inputs to same Q,K,V
        enc_outputs = self.pos_ffn(
            enc_outputs
        )  # enc_outputs: [batch_size x len_q x d_model]

        return enc_outputs, attn


class CCC(torch.nn.Module):
    def __init__(
        self,
        n_layers,
        device,
        d_model,
        d_k,
        d_v,
        d_ff,
        n_heads,
        G,
        pretrained_node_embedding_tensor,
        is_pre_trained,
        base_embedding_dim,
        max_length,
        num_gcn_layers,
        node_edge_composition_func,
        gcn_option,
        get_embeddings,
        ent2id,
        rel2id,
        fine_tuning_layer=False,
    ):

        super().__init__()
        self.n_layers = n_layers
        self.d_model = d_model
        self.max_length = max_length
        self.get_embeddings = get_embeddings
        self.node_edge_composition_func = node_edge_composition_func
        self.gcn_option = gcn_option
        self.fine_tuning_layer = fine_tuning_layer
        self.no_nodes = G.get_number_of_nodes()
        self.device = device
        if self.gcn_option == "preprocess":
            self.num_gcn_layers = num_gcn_layers
        elif self.gcn_option == "alternate":
            assert num_gcn_layers % 2 == 0
            self.num_gcn_layers = int(num_gcn_layers / 2)

        self.gcn_graph_encoder = GCNGraphEncoder(
            device,
            G,
            pretrained_node_embedding_tensor,
            is_pre_trained,
            base_embedding_dim,
            max_length,
            ent2id,
            rel2id,
        )

        if self.gcn_option in ["preprocess", "alternate"]:
            self.gcn_transform = GCNTransform(
                base_embedding_dim, self.num_gcn_layers, self.node_edge_composition_func
            )

        self.layers = torch.nn.ModuleList(
            [EncoderLayer(d_model, d_k, d_v, d_ff, n_heads) for _ in range(n_layers)]
        ).to(device)
        self.linear = torch.nn.Linear(d_model, d_model).to(device)
        self.norm = torch.nn.LayerNorm(d_model).to(device)

        # decoder
        self.decoder = torch.nn.Linear(self.d_model, self.no_nodes).to(device)

    def set_fine_tuning(self):
        self.fine_tuning_layer = True

    def gcn_out2bert_input(self, node_emb, batch_id_maps, input_ids, special_tokens):
        for ii in range(len(input_ids)):
            for jj in range(len(input_ids[ii])):
                if str(input_ids[ii][jj]) in batch_id_maps[ii][0]:
                    gcn_id = batch_id_maps[ii][0][str(input_ids[ii][jj])]
                    tmp_embed = node_emb[ii][gcn_id].unsqueeze(0)
                else:
                    special_id = input_ids[ii][jj]
                    tmp_embed = special_tokens[special_id]["embed"]
                if jj == 0:
                    tmp_out = tmp_embed
                else:
                    tmp_out = torch.cat((tmp_out, tmp_embed), 0)
            tmp_out = tmp_out.unsqueeze(0)
            if ii == 0:
                out = tmp_out
            else:
                out = torch.cat((out, tmp_out), 0)

        return out

    def forward(self, subgraphs_list, all_nodes, masked_pos, masked_nodes):
        (
            norm_subgraph_list,
            node_emb,
            relation_emb,
            batch_id_maps,
            special_tokens_embed,
        ) = self.gcn_graph_encoder(subgraphs_list, masked_nodes)
        if self.gcn_option == "preprocess":
            # Preprocess bert input with 2 or 4 GCN layers
            node_emb, relation_emb = self.gcn_transform(
                "subgraph_list", norm_subgraph_list, node_emb, relation_emb
            )
            # print("Processed GCN subgraph encoder")
        node_emb = self.gcn_out2bert_input(
            node_emb, batch_id_maps, all_nodes, special_tokens_embed
        )
        output = node_emb.to(self.device)

        # print('check1',len(all_nodes[0]))
        enc_self_attn_mask = get_attn_pad_mask(all_nodes, all_nodes, self.no_nodes,self.device)

        for layer in self.layers:
            if self.gcn_option == "alternate":
                # FIXME is this option working?
                # Preprocess with 1/2 GCN layer before each bert layer
                node_emb, relation_emb = self.gcn_transform(
                    "subgraph_list", norm_subgraph_list, output.cpu(), relation_emb
                )
                node_emb = self.gcn_out2bert_input(
                    node_emb, batch_id_maps, all_nodes, special_tokens
                )
                output = node_emb.to(self.device)
            output, enc_self_attn = layer(output, enc_self_attn_mask)
            try:
                layer_output = torch.cat((layer_output, output.unsqueeze(1)), 1)
            except NameError:  # FIXME - replaced bare except
                layer_output = output.unsqueeze(1)

            if self.fine_tuning_layer:
                try:
                    att_output = torch.cat((att_output, enc_self_attn.unsqueeze(0)), 0)
                except NameError: # FIXME - replaced bare except
                    att_output = enc_self_attn.unsqueeze(0)

        # new added for ablation study
        if self.n_layers == 0:
            layer_output = output.unsqueeze(1)
            att_output = "NA"

        if self.fine_tuning_layer:
            # print(output.size(), layer_output.size(), att_output.size())
            return output, layer_output, att_output
        else:
            masked_pos = masked_pos[:, :, None].expand(
                -1, -1, output.size(-1)
            )  # [batch_size, maxlen, d_model]
            h_masked = torch.gather(
                output, 1, masked_pos.to(torch.int64).to(self.device)
            )  # masking position [batch_size, len, d_model]
            h_masked = self.norm(gelu(self.linear(h_masked)))
            pred_score = self.decoder(h_masked)  # [batch_size, maxlen, n_vocab]
            # print('check====', pred_score.size())

            if self.get_embeddings:
                return pred_score, output
            else:
                return pred_score


class FinetuneLayer(torch.nn.Module):
    def __init__(
        self,
        device,
        d_model,
        ft_d_ff,
        ft_layer,
        ft_drop_rate,
        attr_graph,
        ft_input_option,
        n_layers,
    ):

        super().__init__()
        self.d_model = d_model
        self.ft_layer = ft_layer
        self.ft_input_option = ft_input_option
        self.n_layers = n_layers

        if ft_input_option in ["last", "last4_sum"]:
            cnt_layers = 1
        elif ft_input_option in ["last4_cat"]:
            cnt_layers = 4

        if self.n_layers == 0:
            cnt_layers = 1

        if self.ft_layer == "linear":
            self.ft_decoder = torch.nn.Linear(d_model * cnt_layers, d_model).to(device)
            print(self.n_layers, cnt_layers, self.ft_decoder)
            self.dropout = torch.nn.Dropout(ft_drop_rate).to(device)

        elif self.ft_layer == "ffn":
            self.ffn1 = torch.nn.Linear(d_model * cnt_layers, ft_d_ff).to(device)
            print(self.n_layers, cnt_layers, self.ffn1)
            self.dropout = torch.nn.Dropout(ft_drop_rate).to(device)
            self.ffn2 = torch.nn.Linear(ft_d_ff, d_model).to(device)

    def forward(self, graphbert_layer_output):
        """
        graphbert_output = batch_sz * [CLS, source, target, relation, SEP] *
        [emb_size]
        """
        if self.ft_input_option == "last":
            # use the output from laster layer of graphbert
            graphbert_output = graphbert_layer_output[:, -1, :, :].squeeze(1)
            source_embedding = graphbert_output[:, 0, :].unsqueeze(1)
            destination_embedding = graphbert_output[:, 1, :].unsqueeze(1)
        else:
            # concatenate the output from the last four last four layers
            # add for ablation study
            no_layers = graphbert_layer_output.size(1)
            if no_layers == 1:
                start_layer = 0
            else:
                start_layer = no_layers - 4
            for ii in range(start_layer, no_layers):
                source_embed = graphbert_layer_output[:, ii, 0, :].unsqueeze(1)
                destination_embed = graphbert_layer_output[:, ii, 1, :].unsqueeze(1)
                if self.ft_input_option == "last4_cat":
                    try:
                        source_embedding = torch.cat(
                            (source_embedding, source_embed), 2
                        )
                        destination_embedding = torch.cat(
                            (destination_embedding, destination_embed), 2
                        )
                    except:
                        source_embedding = source_embed
                        destination_embedding = destination_embed
                elif self.ft_input_option == "last4_sum":
                    try:
                        source_embedding = torch.add(source_embedding, 1, source_embed)
                        destination_embedding = torch.add(
                            destination_embedding, 1, destination_embed
                        )
                    except:
                        source_embedding = source_embed
                        destination_embedding = destination_embed
        # print(source_embedding.size(), destination_embedding.size())

        if self.ft_layer == "linear":
            src_embedding = self.ft_decoder(source_embedding)
            dst_embedding = self.ft_decoder(destination_embedding)
        elif self.ft_layer == "ffn":
            src_embedding = torch.relu(self.dropout(self.ffn1(source_embedding)))
            src_embedding = self.ffn2(src_embedding)
            dst_embedding = torch.relu(self.dropout(self.ffn1(destination_embedding)))
            dst_embedding = self.ffn2(dst_embedding)

        dst_embedding = dst_embedding.transpose(1, 2)
        pred_score = torch.bmm(src_embedding, dst_embedding).squeeze(1)
        #pred_score = torch.sigmoid(pred_score)
        # print('check+++++', pred_score.size())

        return pred_score, src_embedding, dst_embedding.transpose(1, 2)



import argparse
import json
import random
import os
import time
import shutil
from typing import List
import warnings
warnings.filterwarnings('ignore')
import torch
import torch.nn as nn
import numpy as np
from gensim.models import Word2Vec
import shutil
import pandas as pd
import math
import re
import networkx as nx
from .processing.attributed_graph import AttributedGraph

from .ccc_model.process_walks_gcn import Processing_GCN_Walks
from .finetuning import (
    run_finetuning_wkfl2,
    run_finetuning_wkfl3,
    setup_finetuning_input,
)
import torch.optim as optim
from torch.autograd import Variable
from .pretraining import run_pretraining, setup_pretraining_input
from .processing.context_generator import ContextGenerator
from .processing.generic_attributed_graph import GenericGraph
from .utils.data_utils import load_pretrained_node2vec
from .utils.evaluation import run_evaluation_main
from .utils.link_predict import find_optimal_cutoff, link_prediction_eval
from .utils.utils import get_id_map, load_pickle
from .ccc_model.ccc_model import CCC, FinetuneLayer
from .ccc_model.process_walks_gcn import Processing_GCN_Walks
from .utils.context_metrics import PathMetrics
from .utils.data_utils import create_finetune_batches2
from .utils.utils import load_pickle, show_progress


def get_set_embeddings_details(args):
    if not args.pretrained_embeddings:
        if args.pretrained_method == "compgcn":
            args.pretrained_embeddings = (
                f"{args.emb_dir}/{args.data_name}/"
                f"act_{args.data_name}_{args.node_edge_composition_func}_500.out"
            )
        elif args.pretrained_method == "node2vec":
            args.base_embedding_dim = 128
            args.pretrained_embeddings = (
                f"{args.emb_dir}/{args.data_name}/{args.data_name}.emd"
            )
    else:
        if args.pretrained_method == "compgcn":
            args.base_embedding_dim = 200
        elif args.pretrained_method == "node2vec":
            args.base_embedding_dim = 128
    return args.pretrained_embeddings, args.base_embedding_dim


def get_graph(args, data_path, false_edge_gen):
    print("\n Loading graph...")
    attr_graph = GenericGraph(data_path, false_edge_gen)
    context_gen = ContextGenerator(attr_graph, args.num_walks_per_node)

    return attr_graph, context_gen


def get_test_edges(paths: List[str], sep: str):
    # edges = set()
    edges = []
    for path in paths:
        with open(path, "r") as f:
            for line in f:
                tokens = line.strip().split(sep)
                etype = tokens[0]
                source = tokens[1]
                destination = tokens[2]
                label = tokens[3]
                edge = (etype, source, destination, label)
                # edges.add(edge)
                edges.append(edge)

    return edges


class GenericGraph_pre(AttributedGraph):
    def __init__(
        self, main_dir, false_edge_gen, attributes_file="", sample_training_set=False
    ):
        """
        Assumes derived classes will create a networkx graph object along with
        following
        1) self.unique_relations = set()
        2) self.node_attr_dfs = dict()
        3) self.node_types = dict()
        4) self.G  = nx.Graph()
        """
        super().__init__()
        self.main_dir = main_dir
        self.false_edge_gen = false_edge_gen
        # self.load_graph(main_dir + "/kg_final.txt")
        self.load_graph(main_dir + "/test.txt")
        self.create_normalized_node_id_map()

        # get product attributes
        self.node_attr = dict()
        self.node_attr_list = []

        if attributes_file == "":
            for node_id in self.G.nodes():
                self.node_attr[node_id] = dict()
                self.node_attr[node_id]["defattr"] = "True"
            self.node_attr_list = ["defattr"]
        else:
            raise NotImplementedError

        self.node_attr_dfs = {
            "deftype": pd.DataFrame.from_dict(self.node_attr, orient="index")
        }
        self.sample_training_set = sample_training_set
        self.get_nodeid2rowid()
        self.get_rowid2nodeid()
        self.get_rowid2vocabid_map()
        self.set_relation_id()

    def generate_link_prediction_dataset(self, outdir, fr_test_edges):
        print("In GenericGraph link Prediction Generation")
        false_edge_gen = self.false_edge_gen

        self.test_edges = self.load_test_edges(self.main_dir + "/test.txt")
        return self.test_edges

    def load_graph(self, filename):
        """
        This method should be implemented by all DataLoader classes to fill in
        values for
        1) self.G
        2) self.node_attr_dfs
        3) self.unique_relations
        4) self.node_types
        """
        with open(filename, "r") as f:
            for line in f:
                arr = line.strip().split(" ")
                if len(arr) >= 3:
                    edge_type = arr[0]
                    source = arr[1]
                    target = arr[2]
                    if len(arr) == 3 or (len(arr) == 4 and arr[3] == "1"):
                        self.G.add_edge(source, target, label=edge_type)
                        self.unique_relations.add(edge_type)
                        self.node_types[source] = "deftype"
                        self.node_types[target] = "deftype"
        return

    def load_test_edges(self, filename):
        true_and_false_edges = []
        with open(filename, "r") as f:
            for line in f:
                arr = line.strip().split()
                relation = arr[0]
                source = arr[1]
                target = arr[2]
                if len(arr) == 4:
                    true_and_false_edges.append((relation, source, target, arr[3]))
                else:
                    # GATNE dataset have no false edges for training
                    true_and_false_edges.append((relation, source, target, "1"))
        return true_and_false_edges

    def get_continuous_cols(self):
        return {"deftype": None}

    def get_wide_cols(self):
        return {"deftype": self.node_attr_list}

class ContextGenerator_pre:
    def __init__(
        self,
        attributed_graph_obj,
        num_walks_per_node,
        pretrained_embeddings=None,
        optimal_context_per_relation=None,  # FIXME - removed dangerous empty dict init
    ):
        self.attr_graph = attributed_graph_obj
        self.num_walks_per_node = num_walks_per_node
        self.pretrained_embeddings = pretrained_embeddings
        random.seed(10)

    def get_random_k_nbrs(self, nodeid, exclude_list, k):
        """
        returns randomly sampled 'k' nbrs for a given node as
        list[node_ids]
        if k == -1 :  returns all neighbours
        """
        all_nbrs = list(set(list(nx.all_neighbors(self.attr_graph.G, nodeid))))
        valid_nbrs = all_nbrs
        if exclude_list is None or len(exclude_list) == 0:
            valid_nbrs = all_nbrs
        else:
            valid_nbrs = [x for x in all_nbrs if x not in exclude_list]
        if k >= 0 and len(valid_nbrs) >= k:
            return random.sample(valid_nbrs, k)
        else:
            return all_nbrs

    def gen_random_walk_context(
        self, beam_width, max_num_edges, walk_type, valid_patterns=None
    ):  # FIXME - removed dangerous valid_patterns list init
        """
        Given as input an attributed networkx graph, generates
        random walk based node_context for  each node.
        Returns : list[node_context] == list[list[(source, dest, relation)]]
        Node context is generated by performing random walk around the
        node with beam_width = beam_width and depth = 2
        """
        G = self.attr_graph.get_graph()
        all_contexts = []
        for source in list(G.nodes()):
            # for u,v in nx.bfs_beam_edges(G, source, centrality.get, beam_width):
            # for u,v in nx.bfs_beam_edges(G, source, random_neighbour_goodness, beam_width):
            #    node_context.append((u,v, G.edges[u,v]['label']))
            for _ in range(self.num_walks_per_node):
                if walk_type == "bfs":
                    node_context = self.bfs_beam(source, beam_width)
                elif walk_type == "dfs":
                    node_context = self.random_chain(source, max_num_edges)
                else:
                    print("Unknown context generation strategy, select bfs/dfs")
                all_contexts.append(node_context)
        # valid_patterns = set(['1_1_1', '2_2_2', '1_2_1', '2_1_2'])
        # all_contexts = self.filter_walks_by_patterns(all_contexts, valid_patterns)
        return all_contexts

    def filter_walks_by_patterns(self, walks, valid_patterns):
        # valid_walks = [walk for walk in walks  if self.get_pattern(walk) in
        #               valid_patterns]
        valid_walks = []
        for walk in walks:
            walk_pattern = self.get_pattern(walk)
            if walk_pattern in valid_patterns:
                valid_walks.append(walk)
        return valid_walks

    def get_pattern(self, walk, first_last_same=False):
        """
        given a path = list[(source, target, relation)]
        returns a pattern = [source[0].type, source[1].type, source[2].type...]
        """
        pattern = []
        for edge in walk:
            # source = edge[0]
            relation = edge[2]
            # source_type = self.attr_graph.get_node_type(source)
            # pattern.append(source_type)
            pattern.append(relation)
        return "_".join(pattern)
        # return pattern

    def random_chain(self, source, max_num_edges):
        # generates a random walk of length=max_num_edges, without cycles
        # starting from source
        path = []
        nodes_so_far = []
        G = self.attr_graph.get_graph()
        for _ in range(max_num_edges):
            nbr = self.get_random_k_nbrs(source, exclude_list=nodes_so_far, k=1)[0]
            try:
                relation = G.edges[source, nbr]["label"]
            except KeyError:  # FIXME - replaced bare except
                relation = G.edges[nbr, source]["label"]
            path.append((source, nbr, relation))
            nodes_so_far.append(source)
            source = nbr
        return path

    def bfs_beam(self, source, beam_width):
        """
        Given a source node, a beam width and depth=hops
        return a subgraph around source node .
        The subgraph is retuned as [(source, target, relation)]
        """
        subgraph = list()
        G = self.attr_graph.get_graph()
        source_nbrs = self.get_random_k_nbrs(source, exclude_list=[], k=beam_width)
        for x in source_nbrs:
            try:
                relation = G.edges[source, x]["label"]
            except KeyError:  # FIXME - replaced bare except
                relation = G.edges[x, source]["label"]
            subgraph.append((source, x, relation))
        for src_nbr in source_nbrs:
            src_nbr_hop2_cands = self.get_random_k_nbrs(
                src_nbr, exclude_list=[source], k=beam_width
            )
            for x in src_nbr_hop2_cands:
                try:
                    relation = G.edges[src_nbr, x]["label"]
                except KeyError:  # FIXME - replaced bare except
                    relation = G.edges[x, src_nbr]["label"]
                subgraph.append((src_nbr, x, relation))
        # print(len(subgraph))
        return subgraph

    def get_pretrain_subgraphs(
        self,
        data_path,
        data_name,
        beam_width,
        max_num_edges,
        walk_type,
        mode="pretrain",
    ):
        walk_file = os.path.join(data_path, data_name + "_walks.txt")
        if os.path.exists(walk_file):
            print("\n load walks ...")
            fin = open(walk_file, "r")
            all_walks = []
            for line in fin:
                line = json.loads(line)
                all_walks.append(line)
            fin.close()
        else:
            print("\n generate walks ...")
            # context_gen = ContextGenerator(attr_graph)
            # all_walks = self.gen_random_walk_context(beam_width)
            # for pretraining should we geerate walks without taking out test
            # and validation
            all_walks = self.gen_random_walk_context(
                beam_width, max_num_edges, walk_type
            )
            walk_file = os.path.join(data_path, data_name + "_walks.txt")
            fout = open(walk_file, "w")
            for walk in all_walks:
                json.dump(walk, fout)
                fout.write("\n")
            fout.close()
        print("no. of walks", len(all_walks))
        return all_walks

    def get_finetune_subgraphs(
        self, test_edges, beam_width, max_seq_len, option
    ):

        # print("size of train, valid, test set", len(train_edges),
        #      len(valid_edges), len(test_edges))
        #true_test_edges = [x for x in test_edges if x[3] == "1"]
        #true_valid_edges = [x for x in valid_edges if x[3] == "1"]
        #to_be_removed_edges = true_test_edges + true_valid_edges
        #to_be_removed_source_target_pairs = [(x[1], x[2]) for x in to_be_removed_edges]
        #self.attr_graph.G.remove_edges_from(to_be_removed_source_target_pairs)
        walks_by_task_dict = dict()
        for task, task_edges in zip(
            ["test"], [test_edges]
        ):
            task_subgraphs = []
            for edge in task_edges:
                source = edge[1]
                target = edge[2]
                relation = edge[0]
                is_true = edge[3]
                # source_target_contexts = self.get_path_with_edge_label(source, target,
                #                                      max_seq_len-1)
                source_target_contexts = self.get_selective_context(
                    source, target, beam_width, max_seq_len - 1, option
                )
                if len(source_target_contexts) == 0:
                    first_edge = (source, target, relation, is_true)
                    task_subgraphs.append([first_edge])
                else:
                    for subgraph in source_target_contexts:
                        first_edge = (source, target, relation, is_true)
                        subgraph.insert(0, first_edge)
                        task_subgraphs.append(subgraph)
                walks_by_task_dict[task] = task_subgraphs
        # print(walks_by_task_dict)
        # add edges back to graph, in case attributed graph is used by rest of
        # code
        return walks_by_task_dict

    def get_context(self, source, target, beam_width, max_seq_len):
        all_contexts = []
        # for i in range(self.num_walks_per_node):
        source_context = self.bfs_beam(source, beam_width)[0:max_seq_len]
        # target_context = self.bfs_beam(target, beam_width)[0:max_seq_len]
        all_contexts.append(source_context)
        # all_contexts.append(target_context)
        return all_contexts

    def get_selective_context(
        self, source, target, beam_width, max_seq_len, option, valid_patterns=None
    ):  # FIXME - removed dangerous default list init in valid_patterns

        G = self.attr_graph.G
        paths = []
        if option == "shortest":
            try:
                path_node_list = nx.bidirectional_shortest_path(G, source, target)
                path_with_edge_label = []
                for i in range(len(path_node_list) - 1):
                    u = path_node_list[i]
                    v = path_node_list[i + 1]
                    path_with_edge_label.append((u, v, G.edges[u, v]["label"]))
                paths = [path_with_edge_label[0:max_seq_len]]
            except nx.exception.NetworkXNoPath:
                return []

        elif option == "all":
            paths = self.get_path_with_edge_label(source, target, max_seq_len)
            # print("found all paths", source, target, paths)
        elif option == "pattern":
            paths = self.get_path_with_edge_label(source, target, max_seq_len)
            if len(valid_patterns) > 0:
                paths = self.filter_walks_by_patterns(paths, valid_patterns)
            print("found pattern paths", source, target, paths)
        elif option == "random":
            node_list_paths = self.get_random_paths_between_nodes(
                source, target, beam_width, max_seq_len, currpath=[]
            )
            paths = []
            for node_list_path in node_list_paths:
                path_with_edge_label = []
                node_list_path.insert(0, source)
                for i in range(len(node_list_path) - 1):
                    u = node_list_path[i]
                    v = node_list_path[i + 1]
                    path_with_edge_label.append((u, v, G.edges[u, v]["label"]))
                paths.append(path_with_edge_label)
        elif option == "default":
            paths = self.get_context(source, target, beam_width, max_seq_len)
        return paths

    def get_random_paths_between_nodes(
        self, source, target, beam_width, max_seq_len, currpath
    ):
        """
        Generates random path between source and target , by selecting
        'beam_width' number of neighbors and upto length >= max_seq_len
        """
        if len(currpath) > max_seq_len or source == target:
            return []
        all_paths = []
        exclude_list = []

        # print("In generate paths", source, target, currpath)

        source_nbrs = self.get_random_k_nbrs(source, exclude_list, beam_width)
        # print("selected source nbrs", source_nbrs)
        if target in source_nbrs:
            # print("Found path ending in target", source_nbrs, target)
            path = [target]
            all_paths.append(path)

        for n in source_nbrs:
            if n != target and n not in currpath:
                new_source = n
                new_path = list(currpath)
                new_path.append(new_source)
                nbr_paths = self.get_random_paths_between_nodes(
                    new_source, target, beam_width, max_seq_len, new_path
                )
                new_path = new_path[0:-1]
                for p in nbr_paths:
                    p.insert(0, new_source)
                all_paths.extend(nbr_paths)
        return all_paths

    def get_path_metrics(self, G, source, target, paths):
        """
        Metrics per node pair:
            1) Number of paths connecting the nodes
            2) Number of metapaths
            5) Polysemy behavior : Variation in number of unique relations each node participates in
            6)
        """
        return

    def get_path_with_edge_label(self, source, target, max_seq_len):
        G = self.attr_graph.G
        paths = nx.all_simple_paths(G, source, target, cutoff=max_seq_len)
        paths_with_edge_labels = []
        for path in map(nx.utils.pairwise, paths):
            path_with_edge_label = []
            for u, v in path:
                path_with_edge_label.append((u, v, G.edges[u, v]["label"]))
            paths_with_edge_labels.append(path_with_edge_label)
        return paths_with_edge_labels

def get_graph(args, data_path, false_edge_gen):
    print("\n Loading graph...")
    attr_graph = GenericGraph_pre(data_path, false_edge_gen)
    context_gen = ContextGenerator_pre(attr_graph, args.num_walks_per_node)

    return attr_graph, context_gen

def split_data_pre(data_list, data_path, data_name):
    random.shuffle(data_list)
    data_size = len(data_list)
    data_test = data_list
    tasks = ["test"]
    split_file = {
        task: os.path.join(data_path, data_name + "_walks_" + task + ".txt")
        for task in tasks
    }
    for task in tasks:
        print(task, split_file[task])
        fout = open(split_file[task], "w")
        if task == "train":
            walk_data = data_train
        elif task == "validate":
            walk_data = data_validate
        else:
            walk_data = data_test

        for walk in walk_data:
            json.dump(walk, fout)
            fout.write("\n")
        fout.close()

    return data_test

def create_batches_pre(data_list, data_path, data_name, task, batch_size):
    folder = os.path.join(data_path, task)
    if os.path.exists(folder):
        cnt = math.ceil(len(data_list) / batch_size)
        return cnt
    else:
        os.mkdir(folder)
        random.shuffle(data_list)
        data_size = len(data_list)
        cnt = 0
        data = []
        for ii in range(data_size):
            data.append(data_list[ii])
            if len(data) == batch_size:
                file_name = data_name + "_" + task + "_batch_" + str(cnt) + ".txt"
                fout = open(os.path.join(folder, file_name), "w")
                for jj, _ in enumerate(data):
                    json.dump(data[jj], fout)
                    fout.write("\n")
                fout.close()
                data = []
                cnt += 1

        if len(data) > 0:
            file_name = data_name + "_" + task + "_batch_" + str(cnt) + ".txt"
            fout = open(os.path.join(folder, file_name), "w")
            for jj, _ in enumerate(data):
                json.dump(data[jj], fout)
                fout.write("\n")
            fout.close()
            cnt += 1

        return cnt

def setup_pretraining_input_pre(args, attr_graph, context_gen, data_path):
    # split train/validation/test dataset
    tasks = ["test"]
    print("\n Generating subgraphs for pre-training ...")
    all_walks = context_gen.get_pretrain_subgraphs(
        data_path,
        args.data_name,
        args.beam_width,
        args.max_length,
        args.walk_type,
    )
    #print("\n split data to train/validate/test and save the files ...")
    walk_test = split_data_pre(
        all_walks, data_path, args.data_name
    )
    print(len(walk_test))

    # create batches
    num_batches = {}
    for task in tasks:
        if task == "train":
            walk_data = walk_train
        elif task == "validate":
            walk_data = walk_validate
        else:
            walk_data = walk_test
        cnt = create_batches_pre(
            walk_data, data_path, args.data_name, task, args.batch_size
        )
        num_batches[task] = cnt
    print("number of batches for pre-training: ", num_batches)
    return num_batches

def create_finetune_batches2_pre(finetune_data, finetune_path, data_name, batch_size):

    no_batches = {}
    tasks = ["test"]
    #task_file = {task: os.path.join(finetune_path, task + ".txt") for task in tasks}
    for task in tasks:
        print(task)
        all_task_subgraphs = finetune_data[task]
        tmp_no = create_batches_pre(
            all_task_subgraphs, finetune_path, data_name, task, batch_size
        )
        no_batches[task] = tmp_no
    return no_batches

def setup_finetuning_input_pre(args, attr_graph, context_gen):
    # finetune
    print("\n Generate data for finetuning ...")
    num_batches = {}
    finetune_path = args.outdir + "/finetune/"
    if not os.path.exists(finetune_path):
        os.mkdir(finetune_path)

    if os.path.exists(finetune_path + "/finetune_walks.txt"):
        finetune_walks_per_task = json.load(open(finetune_path + "/finetune_walks.txt"))
    else:
        test_edges = attr_graph.generate_link_prediction_dataset(
            finetune_path, fr_valid_edges=0.1, fr_test_edges=0.1)
        
        finetune_walks_per_task = context_gen.get_finetune_subgraphs(
            test_edges,
            args.beam_width,
            args.max_length,
            args.path_option,
        )
        json.dump(
            finetune_walks_per_task, open(finetune_path + "/finetune_walks.txt", "w")
        )

        print(
            "Number of train, valid, test",
            len(test_edges),
        )
    num_batches = create_finetune_batches2_pre(
        finetune_walks_per_task, finetune_path, args.data_name, args.ft_batch_size
    )
    print("No. of batches for finetuning:", num_batches)

    return num_batches


def Predict(
    args, attr_graph, no_batches, pretrained_node_embedding, ent2id, rel2id
):

    # data_path = args.data_path + args.data_name + "/"
    outdir = args.outdir + "/"
    finetune_path = outdir + "finetune/"
    ft_out_dir = "/mnt/test/a3/TestForCCC/Test/single-cell/data/Sst_Sstr2/output/finetune/results"
    #try:
    #    shutil.rmtree(ft_out_dir)
    #   os.mkdir(ft_out_dir)
    #except:  # FIXME - need to replace bare except
    #   os.mkdir(ft_out_dir)
    relations = attr_graph.relation_to_id
    nodeid2rowid = attr_graph.get_nodeid2rowid()
    walk_processor = Processing_GCN_Walks(
        nodeid2rowid, relations, args.n_pred, args.max_length, args.max_pred
    )

    # process minibatch for finetune(ft)
    print("\n processing minibaches for finetuning:")
    ft_batch_input_file = os.path.join(
        finetune_path, args.data_name + "_ft_batch_input.pickled"
    )
    if os.path.exists(ft_batch_input_file):
        print("loading saved files ...")
        ft_batch_input = load_pickle(ft_batch_input_file)
    else:
        ft_batch_input = {}
        tasks = ["test"]
        for task in tasks:
            print(task)
            ft_batch_input[task] = {}
            for batch_id in range(no_batches[task]):
                (
                    subgraphs,
                    all_nodes,
                    labels,
                ) = walk_processor.process_finetune_minibatch(
                    finetune_path, args.data_name, task, batch_id
                )
                ft_batch_input[task][batch_id] = [subgraphs, all_nodes, labels]
        pickle.dump(ft_batch_input, open(ft_batch_input_file, "wb"))

    if args.is_pre_trained:
        pretrained_node_embedding_tensor = pretrained_node_embedding
    else:
        pretrained_node_embedding_tensor = None

    ccc = CCC(
        args.n_layers,
        args.device,
        args.d_model,
        args.d_k,
        args.d_v,
        args.d_ff,
        args.n_heads,
        attr_graph,
        pretrained_node_embedding_tensor,
        args.is_pre_trained,
        args.base_embedding_dim,
        args.max_length,
        args.num_gcn_layers,
        args.node_edge_composition_func,
        args.gcn_option,
        args.get_bert_encoder_embeddings,
        ent2id,
        rel2id,
    )

    pretrained_node_embedding_tensor = ccc.gcn_graph_encoder.node_embedding

    ft_linear = FinetuneLayer(
        args.device,
        args.d_model,
        args.ft_d_ff,
        args.ft_layer,
        args.ft_drop_rate,
        attr_graph,
        args.ft_input_option,
        args.n_layers,
    )
    

    # run in fine tuning mode
    ccc.set_fine_tuning()

    # testing
    print("Begin Testing")
    ccc.eval()
    ft_linear.eval()

    fl_ = os.path.join(ft_out_dir, "finetune_{}.model".format(9))
    ft_linear.load_state_dict(
        torch.load(fl_, map_location=lambda storage, loc: storage)
    )

    pred_data = []
    true_data = []

    with torch.no_grad():
        att_weights = {}
        final_emd = {}
        for batch_test_id in range(no_batches["test"]):
            if batch_test_id % 100 == 0:
                print("Evaluating test batch {}".format(batch_test_id))
            subgraphs, all_nodes, labels = ft_batch_input["test"][batch_test_id]
            masked_pos = torch.randn(args.batch_size, 2)
            masked_nodes = Variable(
                torch.LongTensor([[] for ii in range(args.ft_batch_size)])
            )
            _, layer_output, att_output = ccc(
                subgraphs, all_nodes, masked_pos, masked_nodes
            )
            score, src_embedding, dst_embedding = ft_linear(layer_output)
            score = score.data.cpu().numpy().tolist()
            labels = labels.data.cpu().numpy().tolist()

            att_weights[batch_test_id] = {
                "attentions": att_output,
                "subgraphs": subgraphs,
                "nodes_sequence": all_nodes,
                "labels": labels,
            }
            final_emd[batch_test_id] = {
                "src_emd": src_embedding,
                "dst_emd": dst_embedding,
                "subgraphs": subgraphs,
                "nodes_sequence": all_nodes,
                "labels": labels,
            }
            path_metrics_pretrained.update_batch(subgraphs, labels)
            path_metrics_graphbert_ft.update_batch_graphbert(
                subgraphs, score, labels, layer_output, all_nodes, "graphbert_ft"
            )

            for ii, _ in enumerate(score):
                pred_data.append(score[ii][0])
                true_data.append(labels[ii][0])
            # print(len(score), len(labels))


    return pred_data, true_data



def run_predict(data_name='Apoe_Grm5',data_path='./Test/single-cell/data/',
    outdir='./Test/single-cell/data/Apoe_Grm5/output',
    pretrained_embeddings='./Test/single-cell/data/data_pca.emd',id_maps_dir='Test/single-cell/data/Sst_Sstr2'):

    torch.manual_seed(1234)
    parser = argparse.ArgumentParser()
    parser.add_argument("--device", default="cuda:0", type=str, help='specify device')
    parser.add_argument("--data_name", default="EFNA5_EPHA2", help="name of the dataset")
    parser.add_argument("--data_path", default="./data/", help="path to dataset")
    parser.add_argument("--outdir", default="./data/EFNA5_EPHA2/output", help="path to output dir")
    parser.add_argument(
        "--pretrained_embeddings", default="./data_pca.emd",help="absolute path to pretrained embeddings"
    )
    parser.add_argument(
        "--pretrained_method", default="node2vec", help="compgcn|node2vec"
    )
    # Walks options
    parser.add_argument(
        "--beam_width",
        default=1,
        type=int,
        help="beam width used for generating random walks",
    )
    parser.add_argument(
        "--num_walks_per_node", default=1, type=int, help="walks per node"
    )
    parser.add_argument("--walk_type", default="dfs", help="walk type bfs/dfs")
    parser.add_argument("--max_length", default=5, type=int, help="max length of walks")
    parser.add_argument(
        "--n_pred", default=1, help="number of tokens masked to be predicted"
    )
    parser.add_argument(
        "--max_pred", default=1, help="max number of tokens masked to be predicted"
    )
    # Pretraining options
    parser.add_argument("--lr", default=0.0001, type=float, help="learning rate")
    parser.add_argument(
        "--n_epochs", default=50, type=int, help="number of epochs for training"
    )
    parser.add_argument(
        "--checkpoint", default=100, type=int, help="checkpoint for validation"
    )
    parser.add_argument(
        "--base_embedding_dim",
        default=200,
        type=int,
        help="dimension of base embedding",
    )
    parser.add_argument(
        "--batch_size",
        default=1024,
        type=int,
        help="number of data sample in each batch",
    )
    parser.add_argument(
        "--emb_dir",
        default="data",
        type=str,
        help="Used to generate embeddings path if --pretrained_embeddings is not set",
    )
    parser.add_argument(
        "--get_bert_encoder_embeddings",
        default=False,
        help="indicate if need to get node vectors from BERT encoder output, save code "
             "commented out in src/pretraining.py",
    )
    # BERT Layer options
    parser.add_argument(
        "--n_layers", default=4, type=int, help="number of encoder layers in bert"
    )
    parser.add_argument(
        "--d_model", default=200, type=int, help="embedding size in bert"
    )
    parser.add_argument("--d_k", default=64, type=int, help="dimension of K(=Q), V")
    parser.add_argument("--d_v", default=64, type=int, help="dimension of K(=Q), V")
    parser.add_argument("--n_heads", default=4, type=int, help="number of head in bert")
    parser.add_argument(
        "--d_ff",
        default=200 * 4,
        type=int,
        help="4*d_model, FeedForward dimension in bert",
    )
    # GCN Layer options
    parser.add_argument(
        "--is_pre_trained",default=False,
        action="store_true",
        help="if there is pretrained node embeddings",
    )
    parser.add_argument(
        "--gcn_option",
        default="no_gcn",
        help="preprocess bert input once or alternate gcn and bert, preprocess|alternate|no_gcn",
    )
    parser.add_argument(
        "--num_gcn_layers", default=4, type=int, help="number of gcn layers before bert"
    )
    parser.add_argument(
        "--node_edge_composition_func",
        default="no_rel",
        help="options for node and edge compostion, sub|circ_conv|mult|no_rel",
    )

    # Finetuning options
    parser.add_argument("--ft_lr", default=0.0001, type=float, help="learning rate")
    parser.add_argument(
        "--ft_batch_size",
        default=1024,
        type=int,
        help="number of data sample in each batch",
    )
    parser.add_argument(
        "--ft_checkpoint", default=1000, type=int, help="checkpoint for validation"
    )
    parser.add_argument(
        "--ft_d_ff", default=512, type=int, help="feedforward dimension in finetuning"
    )
    parser.add_argument(
        "--ft_layer", default="ffn", help="options for finetune layer: linear|ffn"
    )
    parser.add_argument(
        "--ft_drop_rate", default=0.1, type=float, help="dropout rate in finetuning"
    )
    parser.add_argument(
        "--ft_input_option",
        default="last4_cat",
        help="which output layer from graphbert will be used for finetuning, last|last4_cat|last4_sum",
    )
    parser.add_argument(
        "--false_edge_gen",
        default="basic",
        help="false edge generation pattern/double/basic",
    )
    parser.add_argument(
        "--ft_n_epochs", default=10, type=int, help="number of epochs for training"
    )
    parser.add_argument(
        "--path_option",
        default="shortest",
        help="fine tuning context generation: shortest/all/pattern/random",
    )
    args = parser.parse_args(args=[])

    # Default values
    args.pretrained_embeddings, args.base_embedding_dim = get_set_embeddings_details(
        args
    )
    args.d_model = args.base_embedding_dim
    args.d_ff = args.base_embedding_dim
    args.data_name = data_name
    args.data_path = data_path
    args.outdir = outdir
    args.pretrained_embeddings = pretrained_embeddings
    #args.n_epochs = n_epochs
    #args.ft_n_epochs = ft_n_epochs

    data_path = f"{args.data_path}/{args.data_name}"
    attr_graph, context_gen = get_graph(args, data_path, args.false_edge_gen)
    attr_graph.dump_stats()
    stime = time.time()
    #id_maps_dir = 
    ent2id = get_id_map(f"{id_maps_dir}/ent2id.txt")
    rel2id = get_id_map(f"{id_maps_dir}/rel2id.txt")

    pretrained_node_embedding = load_pretrained_node2vec(
            args.pretrained_embeddings, ent2id, args.base_embedding_dim
        )
    outdir = data_path+"/outdir"
    isExists = os.path.exists(outdir)
    if isExists:
        shutil.rmtree(outdir)
        os.makedirs(outdir)
    if not isExists:
        os.makedirs(outdir)
    test_path = data_path + "/test.txt"
    test_edges_paths = [test_path]
    test_edges = list(get_test_edges(test_edges_paths, " "))
    print("No. edges in test data: ", len(test_edges))

    pre_num_batches = setup_pretraining_input_pre(args, attr_graph, context_gen, outdir)
    ft_num_batches = setup_finetuning_input_pre(args, attr_graph, context_gen)

    (
        pred_data_test,
        true_data_test,
    ) = Predict(
        args, attr_graph, ft_num_batches, pretrained_node_embedding, ent2id, rel2id
    )

    y_pred = np.zeros(len(pred_data_test), dtype=np.int32)

    for i, _ in enumerate(pred_data_test):
        if pred_data_test[i] >= 0.5:
            y_pred[i] = 1

    return y_pred


from ..helper import *
from .message_passing import MessagePassing

class CompGCNConvBasis(MessagePassing):
	def __init__(self, in_channels, out_channels, num_rels, num_bases, act=lambda x:x, params=None):
		super(self.__class__, self).__init__()

		self.p 			= params
		self.in_channels	= in_channels
		self.out_channels	= out_channels
		self.num_rels 		= num_rels
		self.num_bases 		= num_bases
		self.act 		= act
		self.device		= None

		self.w_loop		= get_param((in_channels, out_channels));
		self.w_in		= get_param((in_channels, out_channels));
		self.w_out		= get_param((in_channels, out_channels));

		self.rel_basis 		= get_param((self.num_bases, in_channels))
		self.rel_wt 		= get_param((self.num_rels*2, self.num_bases))
		self.w_rel 		= get_param((in_channels, out_channels))
		self.loop_rel 		= get_param((1, in_channels));

		self.drop		= torch.nn.Dropout(self.p.dropout)
		self.bn			= torch.nn.BatchNorm1d(out_channels)

		if self.p.bias: self.register_parameter('bias', Parameter(torch.zeros(out_channels)))

	def forward(self, x, edge_index, edge_type, edge_norm=None, rel_embed=None):
		if self.device is None:
			self.device = edge_index.device

		rel_embed = torch.mm(self.rel_wt, self.rel_basis)
		rel_embed = torch.cat([rel_embed, self.loop_rel], dim=0)

		num_edges = edge_index.size(1) // 2
		num_ent   = x.size(0)

		if not self.cache:
			self.in_norm, self.out_norm, self.in_index, self.out_index, self.in_type, self.out_type, self.loop_index, self.loop_type = None, None, None, None, None, None, None, None

		self.in_index, self.out_index = edge_index[:, :num_edges], edge_index[:, num_edges:]
		self.in_type,  self.out_type  = edge_type[:num_edges], 	 edge_type [num_edges:]

		self.loop_index  = torch.stack([torch.arange(num_ent), torch.arange(num_ent)]).to(self.device)
		self.loop_type   = torch.full((num_ent,), rel_embed.size(0)-1, dtype=torch.long).to(self.device)

		self.in_norm     = self.compute_norm(self.in_index,  num_ent)
		self.out_norm    = self.compute_norm(self.out_index, num_ent)
		
		in_res		= self.propagate('add', self.in_index,   x=x, edge_type=self.in_type,   rel_embed=rel_embed, edge_norm=self.in_norm, 	mode='in')
		loop_res	= self.propagate('add', self.loop_index, x=x, edge_type=self.loop_type, rel_embed=rel_embed, edge_norm=None, 		mode='loop')
		out_res		= self.propagate('add', self.out_index,  x=x, edge_type=self.out_type,  rel_embed=rel_embed, edge_norm=self.out_norm,	mode='out')
		out		= self.drop(in_res)*(1/3) + self.drop(out_res)*(1/3) + loop_res*(1/3)

		if self.p.bias: out = out + self.bias
		if self.b_norm: out = self.bn(out)

		return self.act(out), torch.matmul(rel_embed, self.w_rel)[:-1]

	def rel_transform(self, ent_embed, rel_embed):
		if   self.p.opn == 'corr': 	trans_embed  = ccorr(ent_embed, rel_embed)
		elif self.p.opn == 'sub': 	trans_embed  = ent_embed - rel_embed
		elif self.p.opn == 'mult': 	trans_embed  = ent_embed * rel_embed
		else: raise NotImplementedError

		return trans_embed

	def message(self, x_j, edge_type, rel_embed, edge_norm, mode):
		weight 	= getattr(self, 'w_{}'.format(mode))
		rel_emb = torch.index_select(rel_embed, 0, edge_type)
		xj_rel  = self.rel_transform(x_j, rel_emb)
		out	= torch.mm(xj_rel, weight)

		return out if edge_norm is None else out * edge_norm.view(-1, 1)

	def update(self, aggr_out):
		return aggr_out

	def compute_norm(self, edge_index, num_ent):
		row, col	= edge_index
		edge_weight 	= torch.ones_like(row).float()
		deg		= scatter_add( edge_weight, row, dim=0, dim_size=num_ent)	# Summing number of weights of the edges [Computing out-degree] [Should be equal to in-degree (undireted graph)]
		deg_inv		= deg.pow(-0.5)							# D^{-0.5}
		deg_inv[deg_inv	== float('inf')] = 0
		norm		= deg_inv[row] * edge_weight * deg_inv[col]			# D^{-0.5}

		return norm

	def __repr__(self):
		return '{}({}, {}, num_rels={})'.format(
			self.__class__.__name__, self.in_channels, self.out_channels, self.num_rels)


import pickle
from typing import List

import numpy as np
import pandas as pd
import torch
from sklearn.metrics import (
    auc,
    f1_score,
    precision_recall_curve,
    roc_auc_score,
    roc_curve,
)

from .evaluation import run_evaluation_main


class LinkPrediction(torch.nn.Module):
    """
    Implementation of link prediction strategies based on src
    and dst node embeddings. Ranges from dot product based implementation
    to TransE like approach.
    """

    def __init__(self, opt, ent2id, rel2id, pretrained_node_embeddings):
        """
        Args:
            opt: specify the link prediction strategy.
        """
        super(LinkPrediction, self).__init__()
        self.opt = opt
        self.ent2id = ent2id
        self.pretrained_node_embeddings = pretrained_node_embeddings

    def __predict_via_dot_product(self, edge):
        try:
            source_id = self.ent2id[edge[1]]
            target_id = self.ent2id[edge[2]]
        except KeyError:
            source_id = int(edge[1])
            target_id = int(edge[2])
        source_vec = self.pretrained_node_embeddings[source_id]
        target_vec = self.pretrained_node_embeddings[target_id]
        source_vec = source_vec.unsqueeze(0).unsqueeze(0)
        target_vec = target_vec.unsqueeze(0).unsqueeze(0).transpose(1, 2)
        score = torch.bmm(source_vec, target_vec)
        score = torch.sigmoid(score).data.cpu().numpy().tolist()[0][0][0]

        return score

    def forward(self, triple):
        if self.opt == "dot_product":
            score = self.__predict_via_dot_product(triple)

        return score


def get_node_embeddings(path: str):
    node_embeddings = {}
    header = True
    num_nodes = 0
    size_embedding = 0
    line_count = 0
    with open(path, "r") as f:
        for line in f:
            line_count += 1
            tokens = line.strip().split(" ")
            if header:
                num_nodes = int(tokens[0])
                size_embedding = int(tokens[1])
                print(f"Num nodes: {num_nodes}")
                print(f"Size embedding vector: {size_embedding}")
                header = False
                continue
            node_id = str(tokens[0])
            embedding_vector = np.asarray(tokens[1:], dtype="float32")
            # torch.tensor(np.asarray(tokens[1:], dtype='float32'))
            node_embeddings[node_id] = embedding_vector

    print(f"Num lines: {line_count}")
    return node_embeddings


def get_node_vals(line: str, data_format: str):
    label = None
    tokens = line.strip().split(" ")
    etype = tokens[0]
    node1 = tokens[1]
    node2 = tokens[2]
    # Don't include non-edges for test or valid data
    if data_format in ("valid", "test"):
        label = tokens[3]

    return etype, node1, node2, label


def get_test_edges(paths: List[str]):
    edges = set()
    for path in paths:
        with open(path, "r") as f:
            for line in f:
                etype, source, destination, label = get_node_vals(line, "test")
                edge = (etype, source, destination, label)
                edges.add(edge)
    print(f"Num test edges: {len(edges)}")
    return edges


def get_scores(node_embeddings: dict, test_edges, poly: bool = False, facets: int = 5):
    # Get array of node vector embeddings
    if poly:
        source_embeddings = [node_embeddings[int(edge[1])] for edge in test_edges]
        destination_embeddings = [node_embeddings[int(edge[2])] for edge in test_edges]
    else:
        source_embeddings = [node_embeddings[edge[1]] for edge in test_edges]
        destination_embeddings = [node_embeddings[edge[2]] for edge in test_edges]

    if poly:
        source_embeddings_combo = []
        destination_embeddings_combo = []
        for i, source_embedding in enumerate(source_embeddings):
            destination_embedding = destination_embeddings[i]
            for source_embedding_item in source_embedding:
                for destination_embedding_item in destination_embedding:
                    source_embeddings_combo.append(source_embedding_item)
                    destination_embeddings_combo.append(destination_embedding_item)
        source_embeddings = source_embeddings_combo
        destination_embeddings = destination_embeddings_combo

    # Convert arrays to pytorch tensors
    source_embeddings = torch.Tensor(source_embeddings)
    destination_embeddings = torch.Tensor(destination_embeddings)

    # Get prediction scores
    scores = torch.sigmoid(
        torch.bmm(
            source_embeddings.unsqueeze(dim=1),
            destination_embeddings.unsqueeze(dim=1).transpose(1, 2),
        )
    )
    return scores


def find_optimal_cutoff(target, predicted):
    fpr, tpr, threshold = roc_curve(target, predicted)
    i = np.arange(len(tpr))
    roc = pd.DataFrame(
        {
            "tf": pd.Series(tpr - (1 - fpr), index=i),
            "threshold": pd.Series(threshold, index=i),
        }
    )
    roc_t = roc.loc[(roc.tf - 0).abs().argsort()[:1]]

    return list(roc_t["threshold"])


def test_link_prediction(
    test_edges: list, node_embeddings: dict, poly: bool = False, facets: int = 1
):
    true_list = []
    prediction_list = []
    true_num = 0
    n_facet_combos = facets ** 2

    labels = [int(edge[3]) for edge in test_edges]
    true_edge_indexes = [i for i, label in enumerate(labels) if label == 1]
    false_edge_indexes = [i for i, label in enumerate(labels) if label == 0]
    scores = get_scores(node_embeddings, test_edges, poly)

    print(f"Num true edges: {len(true_edge_indexes)}")
    for ti in true_edge_indexes:
        start = ti * n_facet_combos
        end = start + n_facet_combos
        tmp_score = float(max(scores[start:end]))
        true_list.append(1)
        prediction_list.append(tmp_score)
        true_num += 1

    print(f"Num false edges: {len(false_edge_indexes)}")
    for fi in false_edge_indexes:
        start = fi * n_facet_combos
        end = start + n_facet_combos
        tmp_score = float(max(scores[start:end]))
        true_list.append(0)
        prediction_list.append(tmp_score)

    sorted_pred = prediction_list[:]
    sorted_pred.sort()

    y_true = np.array(true_list)
    threshold = find_optimal_cutoff(y_true, prediction_list)[0]
    # This was the original threshold approximator
    # threshold2 = sorted_pred[-true_num]

    y_pred = np.zeros(len(prediction_list), dtype=np.int32)
    for i, _ in enumerate(prediction_list):
        if prediction_list[i] >= threshold:
            y_pred[i] = 1

    print(f"y_true: {y_true}")
    y_scores = np.array(prediction_list)
    ps, rs, _ = precision_recall_curve(y_true, y_scores)
    # Note we only have one edge_type so we don't have to take the mean
    roc_auc = roc_auc_score(y_true, y_scores)
    pr_auc = auc(rs, ps)
    f1 = f1_score(y_true, y_pred)
    print(f"ROC-AUC: {roc_auc}")
    print(f"PR-AUC: {pr_auc}")
    print(f"F1: {f1}")


def get_poly_embeddings(path: str):
    with open(path, "rb") as f:
        embeddings = pickle.load(f)
    return embeddings


def avg_poly_embeddings(poly_embeddings):
    poly_embeddings_shape = poly_embeddings.shape
    num_samples = poly_embeddings_shape[0]
    embedding_size = poly_embeddings_shape[2]
    avg_poly_embeddings = np.ndarray((num_samples, embedding_size))
    for n in range(num_samples):
        emb_item = poly_embeddings[n]
        # avg_emb_item = np.average(emb_item, axis=0)
        # min_emb_item = np.amin(emb_item, axis=0)
        # max_emb_item = np.amax(emb_item, axis=0)
        sum_emb_item = np.sum(emb_item, axis=0)
        avg_poly_embeddings[n] = sum_emb_item
    print()
    return avg_poly_embeddings


def link_prediction_eval(
    valid_edges, test_edges, ent2id, rel2id, pretrained_node_embeddings
):
    """
    Args:
        test_edges (List[triple]): list of edges loaded from test.txt
        node_embedding (NodeEmbedding): class implementing NodeEmbedding
                interface

    Returns:
        Precision, Recall, F-Score for link prediction test
    """
    link_prediction = LinkPrediction(
        "dot_product", ent2id, rel2id, pretrained_node_embeddings
    )

    pred_data = {}
    true_data = {}
    for data_type in ["valid", "test"]:
        pred_data[data_type] = []
        true_data[data_type] = []
        if data_type == "valid":
            edges = valid_edges
        else:
            edges = test_edges
        for edge in edges:
            label = edge[3]
            true_data[data_type].append(int(label))
            score = link_prediction(edge)
            pred_data[data_type].append(score)

    valid_true_data = np.array(true_data["valid"])
    threshold = find_optimal_cutoff(valid_true_data, pred_data["valid"])[0]
    run_evaluation_main(
        test_edges, pred_data["test"], true_data["test"], threshold, header="workflow1"
    )



import random

import pandas as pd
from .attributed_graph import AttributedGraph


class GenericGraph(AttributedGraph):
    def __init__(
        self, main_dir, false_edge_gen, attributes_file="", sample_training_set=False
    ):
        """
        Assumes derived classes will create a networkx graph object along with
        following
        1) self.unique_relations = set()
        2) self.node_attr_dfs = dict()
        3) self.node_types = dict()
        4) self.G  = nx.Graph()
        """
        super().__init__()
        self.main_dir = main_dir
        self.false_edge_gen = false_edge_gen
        # self.load_graph(main_dir + "/kg_final.txt")
        self.load_graph(main_dir + "/train.txt")
        self.load_graph(main_dir + "/valid.txt")
        self.load_graph(main_dir + "/test.txt")
        self.create_normalized_node_id_map()

        # get product attributes
        self.node_attr = dict()
        self.node_attr_list = []

        if attributes_file == "":
            for node_id in self.G.nodes():
                self.node_attr[node_id] = dict()
                self.node_attr[node_id]["defattr"] = "True"
            self.node_attr_list = ["defattr"]
        else:
            raise NotImplementedError

        self.node_attr_dfs = {
            "deftype": pd.DataFrame.from_dict(self.node_attr, orient="index")
        }
        self.sample_training_set = sample_training_set
        self.get_nodeid2rowid()
        self.get_rowid2nodeid()
        self.get_rowid2vocabid_map()
        self.set_relation_id()

    def generate_link_prediction_dataset(self, outdir, fr_valid_edges, fr_test_edges):
        print("In GenericGraph link Prediction Generation")
        false_edge_gen = self.false_edge_gen
        self.train_edges = self.load_test_edges(self.main_dir + "/train.txt")
        
        num_positive_edges = len(self.train_edges)
        print(num_positive_edges)
        if self.sample_training_set:
            self.train_edges = random.sample(self.train_edges, num_positive_edges / 2)
            num_positive_edges = len(self.train_edges)
        if false_edge_gen == "pattern":
            print("Generating false edges by pattern")
            self.train_edges = self.generate_false_edges3(self.train_edges)
        elif false_edge_gen == "double":
            print("Generating false edges by random false target and source selection")
            self.train_edges = self.generate_false_edges2(self.train_edges)
        elif false_edge_gen == "basic":
            print("Generating false edges by random false target")
            self.train_edges = self.generate_false_edges(self.train_edges)
        else:
            neg_edges = self.load_test_edges(self.main_dir + "/negtive.txt")
            self.train_edges = self.generate_false_edges1(self.train_edges,neg_edges)

            #print("Generating false edges by random false target")
            #self.train_edges = self.load_test_edges(self.main_dir + "/train.txt")
        #print(
        #    "Positive training edge, total number of training edges",
        #    num_positive_edges,
        #    len(self.train_edges),
        #)
        #json.dump(self.train_edges, open("train_edges.json", "w"))

        self.valid_edges = self.load_test_edges(self.main_dir + "/valid.txt")
        self.test_edges = self.load_test_edges(self.main_dir + "/test.txt")
        return self.train_edges, self.valid_edges, self.test_edges

    def load_graph(self, filename):
        """
        This method should be implemented by all DataLoader classes to fill in
        values for
        1) self.G
        2) self.node_attr_dfs
        3) self.unique_relations
        4) self.node_types
        """
        with open(filename, "r") as f:
            for line in f:
                arr = line.strip().split(" ")
                if len(arr) >= 3:
                    edge_type = arr[0]
                    source = arr[1]
                    target = arr[2]
                    if len(arr) == 3 or (len(arr) == 4 and arr[3] == "1"):
                        self.G.add_edge(source, target, label=edge_type)
                        self.unique_relations.add(edge_type)
                        self.node_types[source] = "deftype"
                        self.node_types[target] = "deftype"
        return

    """
    def load_node_attributes(self, filename):
        node_attr_list = []
        with open(filename, "r") as f:
            line = f.readline()
            for line in f:
                arr = line.strip().split()
                graph_node_id = arr[0]
                node_attr = [float(x) for x in arr[1:]]
                for i in range(len(node_attr)):
                    self.node_attr[graph_node_id] = dict()
                    node_attr_key = "attr_"+str(i)
                    self.node_attr[graph_node_id][node_attr_key] = node_attr[i]
                    node_attr_list.append(node_attr_key)
        return self.node_attr, list(set(node_attr_list))
    """

    def load_test_edges(self, filename):
        true_and_false_edges = []
        with open(filename, "r") as f:
            for line in f:
                arr = line.strip().split()
                relation = arr[0]
                source = arr[1]
                target = arr[2]
                if len(arr) == 4:
                    true_and_false_edges.append((relation, source, target, arr[3]))
                else:
                    # GATNE dataset have no false edges for training
                    true_and_false_edges.append((relation, source, target, "1"))
        return true_and_false_edges

    def get_continuous_cols(self):
        return {"deftype": None}

    def get_wide_cols(self):
        return {"deftype": self.node_attr_list}



import inspect, torch, torch_scatter

def scatter_(name, src, index, dim_size=None):
	r"""Aggregates all values from the :attr:`src` tensor at the indices
	specified in the :attr:`index` tensor along the first dimension.
	If multiple indices reference the same location, their contributions
	are aggregated according to :attr:`name` (either :obj:`"add"`,
	:obj:`"mean"` or :obj:`"max"`).

	Args:
		name (string): The aggregation to use (:obj:`"add"`, :obj:`"mean"`,
			:obj:`"max"`).
		src (Tensor): The source tensor.
		index (LongTensor): The indices of elements to scatter.
		dim_size (int, optional): Automatically create output tensor with size
			:attr:`dim_size` in the first dimension. If set to :attr:`None`, a
			minimal sized output tensor is returned. (default: :obj:`None`)

	:rtype: :class:`Tensor`
	"""

	assert name in ['add', 'mean', 'max']

	op		= getattr(torch_scatter, 'scatter_{}'.format(name))
	fill_value	= -1e38 if name == 'max' else 0
	out		= op(src, index, 0, None, dim_size) #, fill_value)
	if isinstance(out, tuple):
		out = out[0]

	if name == 'max':
		out[out == fill_value] = 0

	return out


class MessagePassing(torch.nn.Module):
	r"""Base class for creating message passing layers

	.. math::
		\mathbf{x}_i^{\prime} = \gamma_{\mathbf{\Theta}} \left( \mathbf{x}_i,
		\square_{j \in \mathcal{N}(i)} \, \phi_{\mathbf{\Theta}}
		\left(\mathbf{x}_i, \mathbf{x}_j,\mathbf{e}_{i,j}\right) \right),

	where :math:`\square` denotes a differentiable, permutation invariant
	function, *e.g.*, sum, mean or max, and :math:`\gamma_{\mathbf{\Theta}}`
	and :math:`\phi_{\mathbf{\Theta}}` denote differentiable functions such as
	MLPs.
	See `here <https://rusty1s.github.io/pytorch_geometric/build/html/notes/
	create_gnn.html>`__ for the accompanying tutorial.

	"""

	def __init__(self, aggr='add'):
		super(MessagePassing, self).__init__()

		self.message_args = inspect.getargspec(self.message)[0][1:]	# In the defined message function: get the list of arguments as list of string| For eg. in rgcn this will be ['x_j', 'edge_type', 'edge_norm'] (arguments of message function)
		self.update_args  = inspect.getargspec(self.update)[0][2:]	# Same for update function starting from 3rd argument | first=self, second=out

	def propagate(self, aggr, edge_index, **kwargs):
		r"""The initial call to start propagating messages.
		Takes in an aggregation scheme (:obj:`"add"`, :obj:`"mean"` or
		:obj:`"max"`), the edge indices, and all additional data which is
		needed to construct messages and to update node embeddings."""

		assert aggr in ['add', 'mean', 'max']
		kwargs['edge_index'] = edge_index


		size = None
		message_args = []
		for arg in self.message_args:
			if arg[-2:] == '_i':					# If arguments ends with _i then include indic
				tmp  = kwargs[arg[:-2]]				# Take the front part of the variable | Mostly it will be 'x', 
				size = tmp.size(0)
				message_args.append(tmp[edge_index[0]])		# Lookup for head entities in edges
			elif arg[-2:] == '_j':
				tmp  = kwargs[arg[:-2]]				# tmp = kwargs['x']
				size = tmp.size(0)
				message_args.append(tmp[edge_index[1]])		# Lookup for tail entities in edges
			else:
				message_args.append(kwargs[arg])		# Take things from kwargs

		update_args = [kwargs[arg] for arg in self.update_args]		# Take update args from kwargs

		out = self.message(*message_args)
		out = scatter_(aggr, out, edge_index[0], dim_size=size)		# Aggregated neighbors for each vertex
		out = self.update(out, *update_args)

		return out

	def message(self, x_j):  # pragma: no cover
		r"""Constructs messages in analogy to :math:`\phi_{\mathbf{\Theta}}`
		for each edge in :math:`(i,j) \in \mathcal{E}`.
		Can take any argument which was initially passed to :meth:`propagate`.
		In addition, features can be lifted to the source node :math:`i` and
		target node :math:`j` by appending :obj:`_i` or :obj:`_j` to the
		variable name, *.e.g.* :obj:`x_i` and :obj:`x_j`."""

		return x_j

	def update(self, aggr_out):  # pragma: no cover
		r"""Updates node embeddings in analogy to
		:math:`\gamma_{\mathbf{\Theta}}` for each node
		:math:`i \in \mathcal{V}`.
		Takes in the output of aggregation as first argument and any argument
		which was initially passed to :meth:`propagate`."""

		return aggr_out



from collections import Counter

import numpy as np
from sklearn.metrics import (
    accuracy_score,
    auc,
    f1_score,
    mean_squared_error,
    precision_recall_curve,
    precision_recall_fscore_support,
    roc_auc_score,
)


def run_evaluation(pred_data, true_data):
    """For evaluation."""
    pred_data = np.array(pred_data)
    true_data = np.array(true_data)

    print("pred and true label size", pred_data.shape, true_data.shape)
    pred_seq = []
    true_seq = []
    for ii in range(len(pred_data)):
        pred_seq += [pred_data[ii][jj] for jj in range(len(pred_data[ii]))]
        true_seq += [true_data[ii][jj] for jj in range(len(true_data[ii]))]
    print("pred \n", len(pred_seq))  # , pred_seq)
    # print(pred_seq)
    print("true \n", len(true_seq))  # , true_seq)
    # print(true_seq)

    accu = accuracy_score(true_seq, pred_seq)
    mse = mean_squared_error(true_seq, pred_seq)
    accu = np.round(accu, 4)
    mse = np.round(mse, 4)
    print("Accuracy={}, MSE={}".format(accu, mse))
    return accu, mse


def run_evaluation_graph(pred_data, true_data, test_edges, attr_graph):
    """For evaluation."""
    pred_data = np.array(pred_data)
    true_data = np.array(true_data)

    pred_seq = []
    true_seq = []
    count_correct_by_relation_type = Counter()
    count_error_by_relation_type = Counter()

    for ii in range(len(pred_data)):
        pred_seq += [pred_data[ii][jj] for jj in range(len(pred_data[ii]))]
        true_seq += [true_data[ii][jj] for jj in range(len(true_data[ii]))]

    print("pred \n", len(pred_seq))  # , pred_seq)
    # print(pred_seq)
    print("true \n", len(true_seq))  # , true_seq)
    # print(true_seq)
    # print(test_edges[0][0])
    for ii, subgraph in enumerate(test_edges):
        test_edge = subgraph[0]
        # print(test_edge)
        test_relation = test_edge[1]
        if pred_seq[ii] == true_seq[ii]:
            count_correct_by_relation_type[test_relation] += 1
        else:
            count_error_by_relation_type[test_relation] += 1

    print("Correct counts by relation", count_correct_by_relation_type)
    print("Incorrect counts by relation", count_error_by_relation_type)

    print(
        "Precision , recall, fscore-micro",
        precision_recall_fscore_support(true_seq, pred_seq, average="micro"),
    )
    print(
        "Precision , recall, fscore-macro",
        precision_recall_fscore_support(true_seq, pred_seq, average="macro"),
    )

    print(
        "Precision , recall, fscore-weighted",
        precision_recall_fscore_support(true_seq, pred_seq, average="weighted"),
    )

    accu = accuracy_score(true_seq, pred_seq)
    mse = mean_squared_error(true_seq, pred_seq)
    accu = np.round(accu, 4)
    mse = np.round(mse, 4)

    print("Accuracy={}, MSE={}".format(accu, mse))

    return accu, mse


def run_evaluation_main(test_edges, prediction_data, true_data, threshold, header):
    sorted_pred = prediction_data[:]
    sorted_pred.sort()
    # threshold = sorted_pred[-true_num]
    y_pred = np.zeros(len(prediction_data), dtype=np.int32)

    for i, _ in enumerate(prediction_data):
        if prediction_data[i] >= threshold:
            y_pred[i] = 1

    y_true = np.array(true_data)
    y_scores = np.array(prediction_data)

    ps, rs, _ = precision_recall_curve(y_true, y_scores)

    print("----------------------In run_evaluation_main()------------")
    print(
        f"y_true.shape: {y_true.shape}, y_scores.shape: {y_scores.shape}"
        f", y_pred.shape: {y_pred.shape}"
    )
    try:
        roc_auc = roc_auc_score(y_true, y_scores)
    except ValueError:
        roc_auc = 'UNDEFINED'
    f1 = f1_score(y_true, y_pred)
    auc_value = auc(rs,ps)
    print(
        f"{header} : ROC-AUC: {roc_auc},"
        f" F1: {f1}, AUC: {auc_value}"
    )
    print("\nEvaluation by edge type: ")

    y_pred_dict = {}
    y_true_dict = {}
    y_score_dict = {}
    # print(len(prediction_data), len(true_data), len(test_edges))
    for ii, _ in enumerate(test_edges):
        edge_type = test_edges[ii][0]
        try:
            y_pred_dict[edge_type].append(y_pred[ii])
            y_true_dict[edge_type].append(y_true[ii])
            y_score_dict[edge_type].append(y_scores[ii])
        except KeyError:
            y_pred_dict[edge_type] = [y_pred[ii]]
            y_true_dict[edge_type] = [y_true[ii]]
            y_score_dict[edge_type] = [y_scores[ii]]

    for itm in y_pred_dict:
        print("Performance on edge type: ", itm)
        y_true_dict[itm] = np.array(y_true_dict[itm])
        y_pred_dict[itm] = np.array(y_pred_dict[itm])
        y_score_dict[itm] = np.array(y_score_dict[itm])
        ps, rs, _ = precision_recall_curve(y_true_dict[itm], y_score_dict[itm])
        print(
            f"y_true.shape: {y_true_dict[itm].shape},"
            f" y_scores.shape: {y_score_dict[itm].shape},"
            f" y_pred.shape: {y_pred_dict[itm].shape}"
        )
        try:
            roc_auc = roc_auc_score(y_true_dict[itm], y_score_dict[itm])
        except (KeyError, ValueError):
            roc_auc = 'UNDEFINED'
        f1 = f1_score(y_true_dict[itm], y_pred_dict[itm])
        auc_value = auc(rs,ps)
        print(
                f"{header} : ROC-AUC: "
                f"{roc_auc},"
                f" F1: {f1},"
                f" PR-AUC: {auc_value}"
            )
    print("END")
    return y_true,y_scores

